import warnings

import geopandas as gpd
import numpy as np
import pytest
import xarray as xr
from pyproj.crs import CRS
from pyproj.exceptions import CRSError
from shapely.geometry import Point, Polygon

from clisops.core import subset
from clisops.utils.testing import ContextLogger

try:
    import xesmf
except ImportError:
    xesmf = None


class TestSubsetTime:
    nc_poslons = "cmip3/tas.sresb1.giss_model_e_r.run1.atm.da.nc"

    def test_simple(self, nimbus):
        da = xr.open_dataset(nimbus.fetch(self.nc_poslons)).tas
        yr_st = "2050"
        yr_ed = "2059"

        out = subset.subset_time(da, start_date=yr_st, end_date=yr_ed)
        out1 = subset.subset_time(da, start_date=f"{yr_st}-01", end_date=f"{yr_ed}-12")
        out2 = subset.subset_time(
            da, start_date=f"{yr_st}-01-01", end_date=f"{yr_ed}-12-31"
        )
        np.testing.assert_array_equal(out, out1)
        np.testing.assert_array_equal(out, out2)
        np.testing.assert_array_equal(len(np.unique(out.time.dt.year)), 10)
        np.testing.assert_array_equal(out.time.dt.year.max(), int(yr_ed))
        np.testing.assert_array_equal(out.time.dt.year.min(), int(yr_st))

    def test_time_dates_outofbounds(self, caplog, nimbus):
        caplog.set_level("WARNING", logger="clisops")

        da = xr.open_dataset(nimbus.fetch(self.nc_poslons)).tas
        yr_st = "1776"
        yr_ed = "2077"

        with ContextLogger(caplog):
            out = subset.subset_time(
                da, start_date=f"{yr_st}-01", end_date=f"{yr_ed}-01"
            )
        np.testing.assert_array_equal(out.time.dt.year.min(), da.time.dt.year.min())
        np.testing.assert_array_equal(out.time.dt.year.max(), da.time.dt.year.max())

        assert (
            '"start_date" not found within input date time range. Defaulting to minimum time step in xarray object.'
            in caplog.text
        )
        assert (
            '"end_date" not found within input date time range. Defaulting to maximum time step in xarray object.'
            in caplog.text
        )

    def test_warnings(self, caplog, nimbus):
        caplog.set_level("WARNING", logger="clisops")

        da = xr.open_dataset(nimbus.fetch(self.nc_poslons)).tas

        with pytest.raises(ValueError):
            subset.subset_time(da, start_date="2059", end_date="2050")

        with pytest.raises(TypeError):
            subset.subset_time(da, start_yr=2050, end_yr=2059)

        with ContextLogger(caplog):
            subset.subset_time(
                da,
                start_date=2050,  # noqa
                end_date=2055,  # noqa
            )
            assert (
                'start_date and end_date require dates in (type: str) using formats of "%Y", "%Y-%m" or "%Y-%m-%d".'
                in caplog.text
            )

            subset.subset_time(
                da, start_date="2064-01-01T00:00:00", end_date="2065-02-01T03:12:01"
            )

            assert (
                '"start_date" has been nudged to nearest valid time step in xarray object.'
                in caplog.text
            )

            assert (
                '"end_date" has been nudged to nearest valid time step in xarray object.'
                in caplog.text
            )

    def test_time_start_only(self, nimbus):
        da = xr.open_dataset(nimbus.fetch(self.nc_poslons)).tas
        yr_st = "2050"

        # start date only
        with warnings.catch_warnings(record=True) as record:
            out = subset.subset_time(da, start_date=f"{yr_st}-01")
        assert not record

        np.testing.assert_array_equal(out.time.dt.year.min(), int(yr_st))
        np.testing.assert_array_equal(out.time.dt.year.max(), da.time.dt.year.max())

        with warnings.catch_warnings(record=True) as record:
            out = subset.subset_time(da, start_date=f"{yr_st}-07")
        assert not record

        np.testing.assert_array_equal(out.time.dt.year.min(), int(yr_st))
        np.testing.assert_array_equal(out.time.min().dt.month, 7)
        np.testing.assert_array_equal(out.time.dt.year.max(), da.time.dt.year.max())
        np.testing.assert_array_equal(out.time.max(), da.time.max())

        with warnings.catch_warnings(record=True) as record:
            out = subset.subset_time(da, start_date=f"{yr_st}-07-15")
        assert not record
        np.testing.assert_array_equal(out.time.dt.year.min(), int(yr_st))
        np.testing.assert_array_equal(out.time.min().dt.month, 7)
        np.testing.assert_array_equal(out.time.min().dt.day, 15)
        np.testing.assert_array_equal(out.time.dt.year.max(), da.time.dt.year.max())
        np.testing.assert_array_equal(out.time.max(), da.time.max())

    def test_time_end_only(self, nimbus):
        da = xr.open_dataset(nimbus.fetch(self.nc_poslons)).tas
        yr_ed = "2059"

        # end date only
        with warnings.catch_warnings(record=True) as record:
            out = subset.subset_time(da, end_date=f"{yr_ed}-01")
        assert not record

        np.testing.assert_array_equal(out.time.dt.year.max(), int(yr_ed))
        np.testing.assert_array_equal(out.time.max().dt.month, 1)
        np.testing.assert_array_equal(out.time.max().dt.day, 31)
        np.testing.assert_array_equal(out.time.min(), da.time.min())

        with warnings.catch_warnings(record=True) as record:
            out = subset.subset_time(da, end_date=f"{yr_ed}-06-15")
        assert not record

        np.testing.assert_array_equal(out.time.dt.year.max(), int(yr_ed))
        np.testing.assert_array_equal(out.time.max().dt.month, 6)
        np.testing.assert_array_equal(out.time.max().dt.day, 15)
        np.testing.assert_array_equal(out.time.min(), da.time.min())

    def test_time_incomplete_years(self, nimbus):
        da = xr.open_dataset(nimbus.fetch(self.nc_poslons)).tas
        yr_st = "2050"
        yr_ed = "2059"

        out = subset.subset_time(
            da, start_date=f"{yr_st}-07-01", end_date=f"{yr_ed}-06-30"
        )
        out1 = subset.subset_time(da, start_date=f"{yr_st}-07", end_date=f"{yr_ed}-06")
        np.testing.assert_array_equal(out, out1)
        np.testing.assert_array_equal(out.time.dt.year.min(), int(yr_st))
        np.testing.assert_array_equal(out.time.min().dt.month, 7)
        np.testing.assert_array_equal(out.time.min().dt.day, 1)
        np.testing.assert_array_equal(out.time.dt.year.max(), int(yr_ed))
        np.testing.assert_array_equal(out.time.max().dt.month, 6)
        np.testing.assert_array_equal(out.time.max().dt.day, 30)


class TestSubsetGridPoint:
    nc_poslons = "cmip3/tas.sresb1.giss_model_e_r.run1.atm.da.nc"
    nc_tasmax_file = "NRCANdaily/nrcan_canada_daily_tasmax_1990.nc"
    nc_tasmin_file = "NRCANdaily/nrcan_canada_daily_tasmin_1990.nc"
    nc_2dlonlat = "CRCM5/tasmax_bby_198406_se.nc"

    def test_time_simple(self, nimbus):
        da = xr.open_dataset(nimbus.fetch(self.nc_poslons)).tas
        da = da.assign_coords(lon=(da.lon - 360))
        lon = -72.4
        lat = 46.1
        yr_st = "2050"
        yr_ed = "2059"

        out = subset.subset_gridpoint(
            da, lon=lon, lat=lat, start_date=yr_st, end_date=yr_ed
        )
        np.testing.assert_almost_equal(out.lon, lon, 1)
        np.testing.assert_almost_equal(out.lat, lat, 1)
        np.testing.assert_array_equal(len(np.unique(out.time.dt.year)), 10)
        np.testing.assert_array_equal(out.time.dt.year.max(), int(yr_ed))
        np.testing.assert_array_equal(out.time.dt.year.min(), int(yr_st))

    def test_dataset(self, nimbus):
        da = xr.open_mfdataset(
            [nimbus.fetch(self.nc_tasmax_file), nimbus.fetch(self.nc_tasmin_file)],
            combine="by_coords",
        )
        lon = -72.4
        lat = 46.1
        out = subset.subset_gridpoint(da, lon=lon, lat=lat)
        np.testing.assert_almost_equal(out.lon, lon, 1)
        np.testing.assert_almost_equal(out.lat, lat, 1)
        np.testing.assert_array_equal(out.tasmin.shape, out.tasmax.shape)

    @pytest.mark.parametrize(
        "lon,lat", [([-72.4], [46.1]), ([-67.4, -67.3], [43.1, 46.1])]
    )
    @pytest.mark.parametrize("add_distance", [True, False])
    def test_simple(self, lat, lon, add_distance, nimbus):
        da = xr.open_dataset(nimbus.fetch(self.nc_tasmax_file)).tasmax

        out = subset.subset_gridpoint(da, lon=lon, lat=lat, add_distance=add_distance)
        np.testing.assert_almost_equal(out.lon, lon, 1)
        np.testing.assert_almost_equal(out.lat, lat, 1)

        assert ("site" in out.dims) ^ (len(lat) == 1)
        assert ("distance" in out.coords) ^ (not add_distance)

    def test_irregular(self, nimbus):
        da = xr.open_dataset(nimbus.fetch(self.nc_2dlonlat)).tasmax
        lon = -72.4
        lat = 46.1
        out = subset.subset_gridpoint(da, lon=lon, lat=lat)
        np.testing.assert_almost_equal(out.lon, lon, 1)
        np.testing.assert_almost_equal(out.lat, lat, 1)
        assert "site" not in out.dims

        lon = [-72.4, -67.1]
        lat = [46.1, 48.2]
        out = subset.subset_gridpoint(da, lon=lon, lat=lat)
        np.testing.assert_almost_equal(out.lon, lon, 1)
        np.testing.assert_almost_equal(out.lat, lat, 1)
        assert "site" in out.dims

        # dask for lon lat
        da.lon.chunk({"rlon": 10})
        da.lat.chunk({"rlon": 10})
        out = subset.subset_gridpoint(da, lon=lon, lat=lat)
        np.testing.assert_almost_equal(out.lon, lon, 1)
        np.testing.assert_almost_equal(out.lat, lat, 1)

        # test_irregular transposed:
        da1 = xr.open_dataset(nimbus.fetch(self.nc_2dlonlat)).tasmax
        dims = list(da1.dims)
        dims.reverse()
        daT = xr.DataArray(np.transpose(da1.values), dims=dims)
        for d in daT.dims:
            args = dict()
            args[d] = da1[d]
            daT = daT.assign_coords(**args)
        daT = daT.assign_coords(lon=(["rlon", "rlat"], np.transpose(da1.lon.values)))
        daT = daT.assign_coords(lat=(["rlon", "rlat"], np.transpose(da1.lat.values)))

        out1 = subset.subset_gridpoint(daT, lon=lon, lat=lat)
        np.testing.assert_almost_equal(out1.lon, lon, 1)
        np.testing.assert_almost_equal(out1.lat, lat, 1)
        np.testing.assert_array_equal(out, out1)

        # Dataset with tasmax, lon and lat as data variables (i.e. lon, lat not coords of tasmax)
        daT1 = xr.DataArray(np.transpose(da1.values), dims=dims)
        for d in daT1.dims:
            args = dict()
            args[d] = da1[d]
            daT1 = daT1.assign_coords(**args)
        dsT = xr.Dataset(data_vars=None, coords=daT1.coords)
        dsT["tasmax"] = daT1
        dsT["lon"] = xr.DataArray(np.transpose(da1.lon.values), dims=["rlon", "rlat"])
        dsT["lat"] = xr.DataArray(np.transpose(da1.lat.values), dims=["rlon", "rlat"])
        out2 = subset.subset_gridpoint(dsT, lon=lon, lat=lat)
        np.testing.assert_almost_equal(out2.lon, lon, 1)
        np.testing.assert_almost_equal(out2.lat, lat, 1)
        np.testing.assert_array_equal(out, out2.tasmax)

        # Dataset with lon and lat as 1D arrays
        lon = -60
        lat = -45
        da = xr.DataArray(
            np.random.rand(5, 4),
            dims=("time", "site"),
            coords={"time": np.arange(5), "site": np.arange(4)},
        )
        ds = xr.Dataset(
            data_vars={
                "da": da,
                "lon": ("site", np.linspace(lon, lon + 10, 4)),
                "lat": ("site", np.linspace(lat, lat + 5, 4)),
            }
        )
        gp = subset.subset_gridpoint(ds, lon=lon, lat=lat)
        np.testing.assert_almost_equal(gp.lon, lon)
        np.testing.assert_almost_equal(gp.lat, lat)
        assert gp.site == 0

    def test_positive_lons(self, nimbus):
        da = xr.open_dataset(nimbus.fetch(self.nc_poslons)).tas
        lon = -72.4
        lat = 46.1
        out = subset.subset_gridpoint(da, lon=lon, lat=lat)
        np.testing.assert_almost_equal(out.lon, lon + 360, 1)
        np.testing.assert_almost_equal(out.lat, lat, 1)

        out = subset.subset_gridpoint(da, lon=lon + 360, lat=lat)
        np.testing.assert_almost_equal(out.lon, lon + 360, 1)
        np.testing.assert_almost_equal(out.lat, lat, 1)

    def test_raise(self, nimbus):
        da = xr.open_dataset(nimbus.fetch(self.nc_poslons)).tas
        with pytest.raises(ValueError):
            subset.subset_gridpoint(
                da, lon=-72.4, lat=46.1, start_date="2055-03-15", end_date="2055-03-14"
            )
            subset.subset_gridpoint(
                da, lon=-72.4, lat=46.1, start_date="2055", end_date="2052"
            )
        da = xr.open_dataset(nimbus.fetch(self.nc_2dlonlat)).tasmax.drop_vars(
            names=["lon", "lat"]
        )
        with pytest.raises(Exception):
            subset.subset_gridpoint(da, lon=-72.4, lat=46.1)

    def test_tolerance(self, nimbus):
        da = xr.open_dataset(nimbus.fetch(self.nc_poslons)).tas
        lon = -72.5
        lat = 46.2
        out = subset.subset_gridpoint(da, lon=lon, lat=lat, tolerance=1)
        assert out.isnull().all()

        subset.subset_gridpoint(da, lon=lon, lat=lat, tolerance=1e5)


class TestSubsetBbox:
    nc_poslons = "cmip3/tas.sresb1.giss_model_e_r.run1.atm.da.nc"
    nc_tasmax_file = "NRCANdaily/nrcan_canada_daily_tasmax_1990.nc"
    nc_tasmin_file = "NRCANdaily/nrcan_canada_daily_tasmin_1990.nc"
    nc_2dlonlat = "CRCM5/tasmax_bby_198406_se.nc"

    lon = [-75.4, -68]
    lat = [44.1, 47.1]
    lonGCM = [-70.0, -60.0]
    latGCM = [43.0, 59.0]

    @pytest.mark.xfail(
        reason="NetCDF4 has thread safety issues. Consider using h5netcdf in the future",
        strict=False,
    )
    def test_dataset(self, nimbus):
        da = xr.open_mfdataset(
            [nimbus.fetch(self.nc_tasmax_file), nimbus.fetch(self.nc_tasmin_file)],
            combine="by_coords",
        )
        out = subset.subset_bbox(da, lon_bnds=self.lon, lat_bnds=self.lat)
        assert np.all(out.lon >= np.min(self.lon))
        assert np.all(out.lon <= np.max(self.lon))
        assert np.all(out.lat >= np.min(self.lat))
        assert np.all(out.lat <= np.max(self.lat))
        np.testing.assert_array_equal(out.tasmin.shape, out.tasmax.shape)

    def test_simple(self, nimbus):
        da = xr.open_dataset(nimbus.fetch(self.nc_tasmax_file)).tasmax

        out = subset.subset_bbox(da, lon_bnds=self.lon, lat_bnds=self.lat)
        assert out.lon.values.size != 0
        assert out.lat.values.size != 0
        assert np.all(out.lon >= np.min(self.lon))
        assert np.all(out.lon <= np.max(self.lon))
        assert np.all(out.lat.values >= np.min(self.lat))
        assert np.all(out.lat <= np.max(self.lat))

        da = xr.open_dataset(nimbus.fetch(self.nc_poslons)).tas
        da = da.assign_coords(lon=(da.lon - 360))
        yr_st = 2050
        yr_ed = 2059

        out = subset.subset_bbox(
            da,
            lon_bnds=self.lonGCM,
            lat_bnds=self.latGCM,
            start_date=str(yr_st),
            end_date=str(yr_ed),
        )
        assert out.lon.values.size != 0
        assert out.lat.values.size != 0
        assert np.all(out.lon >= np.min(self.lonGCM))
        assert np.all(out.lon <= np.max(self.lonGCM))
        assert np.all(out.lat >= np.min(self.latGCM))
        assert np.all(out.lat <= np.max(self.latGCM))
        np.testing.assert_array_equal(out.time.dt.year.max(), yr_ed)
        np.testing.assert_array_equal(out.time.dt.year.min(), yr_st)

        out = subset.subset_bbox(
            da, lon_bnds=self.lon, lat_bnds=self.lat, start_date=str(yr_st)
        )

        assert out.lon.values.size != 0
        assert out.lat.values.size != 0
        assert np.all(out.lon >= np.min(self.lon))
        assert np.all(out.lon <= np.max(self.lon))
        assert np.all(out.lat >= np.min(self.lat))
        assert np.all(out.lat <= np.max(self.lat))
        np.testing.assert_array_equal(out.time.dt.year.max(), da.time.dt.year.max())
        np.testing.assert_array_equal(out.time.dt.year.min(), yr_st)

        out = subset.subset_bbox(
            da, lon_bnds=self.lon, lat_bnds=self.lat, end_date=str(yr_ed)
        )

        assert out.lon.values.size != 0
        assert out.lat.values.size != 0
        assert np.all(out.lon >= np.min(self.lon))
        assert np.all(out.lon <= np.max(self.lon))
        assert np.all(out.lat >= np.min(self.lat))
        assert np.all(out.lat <= np.max(self.lat))
        np.testing.assert_array_equal(out.time.dt.year.max(), yr_ed)
        np.testing.assert_array_equal(out.time.dt.year.min(), da.time.dt.year.min())

    def test_irregular(self, nimbus):
        da = xr.open_dataset(nimbus.fetch(self.nc_2dlonlat)).tasmax

        out = subset.subset_bbox(da, lon_bnds=self.lon, lat_bnds=self.lat)

        # for irregular lat lon grids data matrix remains rectangular in native proj
        # but with data outside bbox assigned nans.  This means it can have lon and lats outside the bbox.
        # Check only non-nans gridcells using mask
        mask1 = ~(np.isnan(out.sel(time=out.time[0])))
        assert out.lon.values.size != 0
        assert out.lat.values.size != 0
        assert np.all(out.lon.values[mask1.values] >= np.min(self.lon))
        assert np.all(out.lon.values[mask1.values] <= np.max(self.lon))
        assert np.all(out.lat.values[mask1.values] >= np.min(self.lat))
        assert np.all(out.lat.values[mask1.values] <= np.max(self.lat))

    def test_irregular_dataset(self, nimbus):
        da = xr.open_dataset(nimbus.fetch(self.nc_2dlonlat))
        out = subset.subset_bbox(da, lon_bnds=[-150, 100], lat_bnds=[10, 60])
        variables = list(da.data_vars)
        variables.pop(variables.index("tasmax"))
        # only tasmax should be subsetted/masked others should remain untouched
        for v in variables:
            assert out[v].dims == da[v].dims
            np.testing.assert_array_equal(out[v], da[v])

        # ensure results are equal to previous test on DataArray only
        out1 = subset.subset_bbox(da.tasmax, lon_bnds=[-150, 100], lat_bnds=[10, 60])
        np.testing.assert_array_equal(out1, out.tasmax)

        # additional test if dimensions have no coordinates
        da = da.drop_vars(["rlon", "rlat"])
        subset.subset_bbox(da.tasmax, lon_bnds=[-150, 100], lat_bnds=[10, 60])
        # We don't test for equality with previous datasets.
        # Without coords, sel defaults to isel which doesn't include the last element.

    def test_irregular_inverted_dataset(self, nimbus):
        da = xr.open_dataset(nimbus.fetch(self.nc_2dlonlat))
        da_rev = da.sortby("rlat", ascending=False).sortby("rlon", ascending=False)
        out = subset.subset_bbox(da, lon_bnds=[-150, -100], lat_bnds=[10, 60])
        out_rev = subset.subset_bbox(da_rev, lon_bnds=[-150, -100], lat_bnds=[10, 60])
        variables = list(da.data_vars)
        variables.pop(variables.index("tasmax"))
        # only tasmax should be subsetted/masked others should remain untouched
        for v in variables:
            assert out[v].dims == da[v].dims
            np.testing.assert_array_equal(out_rev[v], da[v])

        # ensure results are equal to previous test on DataArray only
        out1 = subset.subset_bbox(da.tasmax, lon_bnds=[-150, -100], lat_bnds=[10, 60])
        out1_rev = subset.subset_bbox(
            da_rev.tasmax, lon_bnds=[-150, -100], lat_bnds=[10, 60]
        )
        diff = (out1 - out1_rev).values
        np.testing.assert_array_equal(np.unique(diff[~np.isnan(diff)]), 0)

        # additional test if dimensions have no coordinates
        da_rev = da_rev.drop_vars(["rlon", "rlat"])
        subset.subset_bbox(da_rev.tasmax, lon_bnds=[-150, -100], lat_bnds=[10, 60])
        # We don't test for equality with previous datasets.
        # Without coords, sel defaults to isel which doesn't include the last element.

    # test datasets with descending coords
    def test_inverted_coords(self):
        lon = np.linspace(-90, -60, 200)
        lat = np.linspace(40, 80, 100)
        da = xr.Dataset(
            data_vars=None, coords={"lon": np.flip(lon), "lat": np.flip(lat)}
        )
        da["data"] = xr.DataArray(
            np.random.rand(lon.size, lat.size), dims=["lon", "lat"]
        )

        out = subset.subset_bbox(da, lon_bnds=self.lon, lat_bnds=self.lat)
        assert out.lon.values.size != 0
        assert out.lat.values.size != 0
        assert np.all(out.lon >= np.min(np.asarray(self.lon)))
        assert np.all(out.lon <= np.max(np.asarray(self.lon)))
        assert np.all(out.lat >= np.min(self.lat))
        assert np.all(out.lat <= np.max(self.lat))

    def test_badly_named_latlons(self, nimbus):
        da = xr.open_dataset(nimbus.fetch(self.nc_tasmax_file))
        extended_latlons = {"lat": "latitude", "lon": "longitude"}
        da_extended_names = da.rename(extended_latlons)
        out = subset.subset_bbox(
            da_extended_names, lon_bnds=self.lon, lat_bnds=self.lat
        )
        assert {"latitude", "longitude"}.issubset(out.dims)

        long_for_some_reason = {"lon": "long"}
        da_long = da.rename(long_for_some_reason)
        out = subset.subset_bbox(da_long, lon_bnds=self.lon, lat_bnds=self.lat)
        assert {"long"}.issubset(out.dims)

        lons_lats = {"lon": "lons", "lat": "lats"}
        da_lonslats = da.rename(lons_lats)
        out = subset.subset_bbox(da_lonslats, lon_bnds=self.lon, lat_bnds=self.lat)
        assert {"lons", "lats"}.issubset(out.dims)

    def test_single_bounds_rectilinear(self, nimbus):
        da = xr.open_dataset(nimbus.fetch(self.nc_tasmax_file)).tasmax

        out = subset.subset_bbox(da, lon_bnds=self.lon)
        assert out.lon.values.size != 0
        assert out.lat.values.size != 0
        np.testing.assert_array_equal(out.lat, da.lat)
        assert np.all(out.lon <= np.max(self.lon))
        assert np.all(out.lon.values >= np.min(self.lon))

        out = subset.subset_bbox(da, lat_bnds=self.lat)
        assert out.lon.values.size != 0
        assert out.lat.values.size != 0
        np.testing.assert_array_equal(out.lon, da.lon)
        assert np.all(out.lat <= np.max(self.lat))
        assert np.all(out.lat.values >= np.min(self.lat))

    def test_single_bounds_curvilinear(self, nimbus):
        da = xr.open_dataset(nimbus.fetch(self.nc_2dlonlat)).tasmax

        out = subset.subset_bbox(da, lon_bnds=self.lon)
        assert out.lon.values.size != 0
        assert out.lat.values.size != 0
        mask1 = ~(np.isnan(out.sel(time=out.time[0])))
        assert np.all(out.lon.values[mask1.values] <= np.max(self.lon))
        assert np.all(out.lon.values[mask1.values] >= np.min(self.lon))

        out = subset.subset_bbox(da, lat_bnds=self.lat)
        assert out.lon.values.size != 0
        assert out.lat.values.size != 0
        mask1 = ~(np.isnan(out.sel(time=out.time[0])))
        assert np.all(out.lat.values[mask1.values] <= np.max(self.lat))
        assert np.all(out.lat.values[mask1.values] >= np.min(self.lat))

    def test_positive_lons(self, nimbus):
        da = xr.open_dataset(nimbus.fetch(self.nc_poslons)).tas

        out = subset.subset_bbox(da, lon_bnds=self.lonGCM, lat_bnds=self.latGCM)
        assert out.lon.values.size != 0
        assert out.lat.values.size != 0
        assert np.all(out.lon >= np.min(np.asarray(self.lonGCM) + 360))
        assert np.all(out.lon <= np.max(np.asarray(self.lonGCM) + 360))
        assert np.all(out.lat >= np.min(self.latGCM))
        assert np.all(out.lat <= np.max(self.latGCM))

        out = subset.subset_bbox(
            da, lon_bnds=np.array(self.lonGCM) + 360, lat_bnds=self.latGCM
        )
        assert np.all(out.lon >= np.min(np.asarray(self.lonGCM) + 360))

    def test_time(self, nimbus):
        da = xr.open_dataset(nimbus.fetch(self.nc_poslons)).tas
        da = da.assign_coords(lon=(da.lon - 360))

        out = subset.subset_bbox(
            da,
            lon_bnds=self.lonGCM,
            lat_bnds=self.latGCM,
            start_date="2050",
            end_date="2059",
        )
        assert out.lon.values.size != 0
        assert out.lat.values.size != 0
        assert np.all(out.lon >= np.min(self.lonGCM))
        assert np.all(out.lon <= np.max(self.lonGCM))
        assert np.all(out.lat >= np.min(self.latGCM))
        assert np.all(out.lat <= np.max(self.latGCM))
        np.testing.assert_array_equal(out.time.min().dt.year, 2050)
        np.testing.assert_array_equal(out.time.min().dt.month, 1)
        np.testing.assert_array_equal(out.time.min().dt.day, 1)
        np.testing.assert_array_equal(out.time.max().dt.year, 2059)
        np.testing.assert_array_equal(out.time.max().dt.month, 12)
        np.testing.assert_array_equal(out.time.max().dt.day, 31)

        out = subset.subset_bbox(
            da,
            lon_bnds=self.lonGCM,
            lat_bnds=self.latGCM,
            start_date="2050-02-05",
            end_date="2059-07-15",
        )
        assert out.lon.values.size != 0
        assert out.lat.values.size != 0
        assert np.all(out.lon >= np.min(self.lonGCM))
        assert np.all(out.lon <= np.max(self.lonGCM))
        assert np.all(out.lat >= np.min(self.latGCM))
        assert np.all(out.lat <= np.max(self.latGCM))
        np.testing.assert_array_equal(out.time.min().dt.year, 2050)
        np.testing.assert_array_equal(out.time.min().dt.month, 2)
        np.testing.assert_array_equal(out.time.min().dt.day, 5)
        np.testing.assert_array_equal(out.time.max().dt.year, 2059)
        np.testing.assert_array_equal(out.time.max().dt.month, 7)
        np.testing.assert_array_equal(out.time.max().dt.day, 15)

    def test_raise(self, nimbus):
        # 1st case
        da = xr.open_dataset(nimbus.fetch(self.nc_poslons)).tas
        with pytest.raises(ValueError):
            subset.subset_bbox(
                da,
                lon_bnds=self.lonGCM,
                lat_bnds=self.latGCM,
                start_date="2056",
                end_date="2055",
            )

        # 2nd case
        da = xr.open_dataset(nimbus.fetch(self.nc_2dlonlat)).tasmax.drop_vars(
            names=["lon", "lat"]
        )
        with pytest.raises(Exception):
            subset.subset_bbox(da, lon_bnds=self.lon, lat_bnds=self.lat)

        # 3rd case
        ds = xr.Dataset(
            data_vars={"var": (("lat", "lon"), np.ones((5, 10)))},
            coords={
                "lat": ("lat", np.zeros(5)),
                "lon": ("lon", np.arange(-10, 0, 1)),
            },
        )
        ds["lat"].attrs["standard_name"] = "latitude"
        ds["lon"].attrs["standard_name"] = "longitude"
        with pytest.raises(ValueError):
            subset.subset_bbox(ds, lon_bnds=(-0.1, 1.0))

    def test_warnings(self, nimbus):
        da = xr.open_dataset(nimbus.fetch(self.nc_poslons)).tas
        da = da.assign_coords(lon=(da.lon - 360))

        with pytest.raises(TypeError):
            subset.subset_bbox(
                da, lon_bnds=self.lon, lat_bnds=self.lat, start_yr=2050, end_yr=2059
            )
        with warnings.catch_warnings(record=True) as record:
            subset.subset_bbox(
                da,
                lon_bnds=self.lon,
                lat_bnds=self.lat,
                start_date="2050",
                end_date="2055",
            )
        assert (
            '"start_yr" and "end_yr" (type: int) are being deprecated. Temporal subsets will soon exclusively'
            ' support "start_date" and "end_date" (type: str) using formats of "%Y", "%Y-%m" or "%Y-%m-%d".'
            not in [str(q.message) for q in record]
        )

    def test_locstream(self):
        da = xr.DataArray(
            [1, 2, 3, 4],
            dims=("site",),
            coords={
                "lat": (("site",), [10, 30, 20, 40]),
                "lon": (("site",), [-50, -80, -70, -100]),
            },
        )
        sub = subset.subset_bbox(da, lon_bnds=[-95, -65], lat_bnds=[15, 35])
        exp = da.isel(site=[1, 2])
        xr.testing.assert_identical(sub, exp)


class TestSubsetShape:
    nc_file = "cmip5/tas_Amon_CanESM2_rcp85_r1i1p1_200701-200712.nc"
    lons_2d_nc_file = "CRCM5/tasmax_bby_198406_se.nc"
    nc_file_neglons = "NRCANdaily/nrcan_canada_daily_tasmax_1990.nc"

    @staticmethod
    def compare_vals(ds, sub, vari, flag_2d=False):
        # Check subsetted values against original.
        imask = np.where(~np.isnan(sub[vari].isel(time=0)))
        if len(imask[0]) > 70:
            rs = np.random.RandomState(42)
            ii = rs.randint(0, len(imask[0]), 70)
        else:
            ii = np.arange(0, len(imask[0]))
        for i in zip(imask[0][ii], imask[1][ii]):
            if flag_2d:
                lat1 = sub.lat[i[0], i[1]]
                lon1 = sub.lon[i[0], i[1]]
                np.testing.assert_array_equal(
                    subset.subset_gridpoint(sub, lon=lon1, lat=lat1)[vari],
                    subset.subset_gridpoint(ds, lon=lon1, lat=lat1)[vari],
                )
            else:
                lat1 = sub.lat.isel(lat=i[0])
                lon1 = sub.lon.isel(lon=i[1])
                # print(lon1.values, lat1.values)
                np.testing.assert_array_equal(
                    sub[vari].sel(lon=lon1, lat=lat1), ds[vari].sel(lon=lon1, lat=lat1)
                )

    def test_wraps(self, tmp_netcdf_filename, nimbus, clisops_test_data):
        ds = xr.open_dataset(nimbus.fetch(self.nc_file))

        # Polygon crosses meridian, a warning should be raised
        with pytest.warns(UserWarning):
            sub = subset.subset_shape(ds, clisops_test_data["meridian_geojson"])

        # No time subsetting should occur.
        assert len(sub.tas) == 12

        # Average temperature at surface for region in January (time=0)
        np.testing.assert_array_almost_equal(
            float(np.mean(sub.tas.isel(time=0))), 285.064, 3
        )
        self.compare_vals(ds, sub, "tas")

        poly = gpd.read_file(clisops_test_data["meridian_multi_geojson"])

        subtas = subset.subset_shape(ds.tas, poly)
        np.testing.assert_array_almost_equal(
            float(np.mean(subtas.isel(time=0))), 281.092, 3
        )

        sub.to_netcdf(tmp_netcdf_filename)
        assert tmp_netcdf_filename.exists()
        with xr.open_dataset(filename_or_obj=tmp_netcdf_filename) as f:
            assert {
                "tas",
            }.issubset(set(f.data_vars))
            subset.subset_shape(ds, clisops_test_data["meridian_multi_geojson"])

    def test_no_wraps(self, tmp_netcdf_filename, nimbus, clisops_test_data):
        ds = xr.open_dataset(nimbus.fetch(self.nc_file))

        with warnings.catch_warnings(record=True) as record:
            sub = subset.subset_shape(ds, clisops_test_data["poslons_geojson"])

        self.compare_vals(ds, sub, "tas")

        # No time subsetting should occur.
        assert len(sub.tas) == 12

        # Average temperature at surface for region in January (time=0)
        np.testing.assert_array_almost_equal(
            float(np.mean(sub.tas.isel(time=0))), 276.732, 3
        )
        # Check that no warnings are raised for meridian crossing
        assert (
            '"Geometry crosses the Greenwich Meridian. Proceeding to split polygon at Greenwich."'
            '" This feature is experimental. Output might not be accurate."'
            not in [str(q.message) for q in record]
        )

        sub.to_netcdf(tmp_netcdf_filename)
        assert tmp_netcdf_filename.exists()
        with xr.open_dataset(filename_or_obj=tmp_netcdf_filename) as f:
            assert {"tas"}.issubset(set(f.data_vars))
            subset.subset_shape(ds, clisops_test_data["poslons_geojson"])

    def test_all_neglons(self, nimbus, clisops_test_data):
        ds = xr.open_dataset(nimbus.fetch(self.nc_file_neglons))

        with warnings.catch_warnings(record=True) as record:
            sub = subset.subset_shape(ds, clisops_test_data["southern_qc_geojson"])

        self.compare_vals(ds, sub, "tasmax")

        # Average temperature at surface for region in January (time=0)
        np.testing.assert_array_almost_equal(
            float(np.mean(sub.tasmax.isel(time=0))), 269.254, 3
        )
        # Check that no warnings are raised for meridian crossing
        assert (
            '"Geometry crosses the Greenwich Meridian. Proceeding to split polygon at Greenwich."'
            '" This feature is experimental. Output might not be accurate."'
            not in [q.message for q in record]
        )

    def test_rotated_pole_with_time(self, nimbus, clisops_test_data):
        ds = xr.open_dataset(nimbus.fetch(self.lons_2d_nc_file))

        with warnings.catch_warnings(record=True) as record:
            sub = subset.subset_shape(
                ds,
                clisops_test_data["eastern_canada_geojson"],
                start_date="1984-06-01",
                end_date="1984-06-15",
            )

        self.compare_vals(
            ds.sel(time=slice("1984-06-01", "1984-06-15")), sub, "tasmax", flag_2d=True
        )

        # Should only have 15 days of data.
        assert len(sub.tasmax) == 15
        # Average max temperature at surface for region on June 1st, 1984 (time=0)
        np.testing.assert_allclose(float(np.mean(sub.tasmax.isel(time=0))), 289.634968)
        # Check that no warnings are raised for meridian crossing
        assert (
            '"Geometry crosses the Greenwich Meridian. Proceeding to split polygon at Greenwich."'
            '" This feature is experimental. Output might not be accurate."'
            not in [str(q.message) for q in record]
        )

    def test_small_poly_buffer(self, tmp_netcdf_filename, nimbus, clisops_test_data):
        ds = xr.open_dataset(nimbus.fetch(self.nc_file))

        with pytest.raises(ValueError):
            subset.subset_shape(ds, clisops_test_data["small_geojson"])

        with pytest.raises(ValueError):
            subset.subset_shape(ds, clisops_test_data["small_geojson"], buffer=0.6)

        sub = subset.subset_shape(ds, clisops_test_data["small_geojson"], buffer=5)
        self.compare_vals(ds, sub, "tas")
        assert len(sub.lon.values) == 3
        assert len(sub.lat.values) == 3

        sub.to_netcdf(tmp_netcdf_filename)
        assert tmp_netcdf_filename.exists()
        with xr.open_dataset(filename_or_obj=tmp_netcdf_filename) as f:
            assert {"tas"}.issubset(set(f.data_vars))

    def test_mask_multiregions(self, nimbus, clisops_test_data):
        ds = xr.open_dataset(nimbus.fetch(self.nc_file))
        regions = gpd.read_file(clisops_test_data["multi_regions_geojson"])
        regions.set_index("id")
        mask = subset.create_mask(
            x_dim=ds.lon, y_dim=ds.lat, poly=regions, wrap_lons=True
        )
        vals, counts = np.unique(mask.values[mask.notnull()], return_counts=True)
        np.testing.assert_array_equal(vals, [0, 1, 2])
        np.testing.assert_array_equal(counts, [58, 250, 22])

    @pytest.mark.skipif(
        xesmf is None, reason="xESMF >= 0.6.2 is needed for average_shape."
    )
    def test_weight_masks_multiregions(self, nimbus, clisops_test_data):
        # rename is due to a small limitation of xESMF 0.5.2
        ds = xr.open_dataset(nimbus.fetch(self.nc_file)).rename(bnds="bounds")
        regions = gpd.read_file(clisops_test_data["multi_regions_geojson"]).set_index(
            "id"
        )
        masks = subset.create_weight_masks(ds, poly=regions)

        np.testing.assert_allclose(masks.sum(["lat", "lon"]), [1, 1, 1])
        np.testing.assert_array_equal(masks.geom.values, regions.index)
        np.testing.assert_allclose(masks.max("geom").sum(), 2.900, 3)

    def test_subset_multiregions(self, nimbus, clisops_test_data):
        ds = xr.open_dataset(nimbus.fetch(self.nc_file))
        regions = gpd.read_file(clisops_test_data["multi_regions_geojson"])
        regions.set_index("id")
        ds_sub = subset.subset_shape(ds, shape=regions)
        assert ds_sub.notnull().sum() == 58 + 250 + 22
        assert ds_sub.tas.shape == (12, 14, 128)

    def test_vectorize_touches_polygons(self):
        """Check that points touching the polygon are included in subset."""
        # Create simple polygon
        poly = Polygon([[0, 0], [1, 0], [1, 1]])
        shape = gpd.GeoDataFrame(geometry=[poly])
        # Create simple data array
        da = xr.DataArray(
            data=[[0, 1], [2, 3]],
            dims=("lon", "lat"),
            coords={"lon": [0, 1], "lat": [0, 1]},
        )
        sub = subset.subset_shape(da, shape=shape)
        assert sub.notnull().sum() == 3

    def test_locstream(self):
        da = xr.DataArray(
            [1, 2, 3, 4, 5, 6, 7, 8, 9],
            dims=("site",),
            coords={
                "lat": (("site",), [55, 55, 55, 40, 40, 40, 25, 25, 25]),
                "lon": (("site",), [-80, -70, -60, -80, -70, -60, -80, -70, -60]),
            },
        )
        poly = Polygon([[-90, 40], [-70, 20], [-50, 40], [-70, 60]])
        shape = gpd.GeoDataFrame(geometry=[poly])
        sub = subset.subset_shape(da, shape=shape)
        exp = da.isel(site=[1, 3, 4, 5, 7])
        xr.testing.assert_identical(sub, exp)

    def test_shapefile_wrapped_wgs84(self, nimbus):
        da = xr.open_dataset(nimbus.fetch(self.nc_file))
        poly_wgs84 = Polygon([[-90, 40], [-70, 20], [-50, 40], [-70, 60]])
        poly_wrapped = Polygon(
            [[-90 + 360, 40], [-70 + 360, 20], [-50 + 360, 40], [-70 + 360, 60]]
        )
        shape = gpd.GeoDataFrame(geometry=[poly_wgs84])
        shape_wrapped = gpd.GeoDataFrame(geometry=[poly_wrapped])
        assert subset.subset_shape(da, shape).equals(
            subset.subset_shape(da, shape_wrapped)
        )

    def test_shapefile_utm(self, nimbus, clisops_test_data):
        da = xr.open_dataset(nimbus.fetch(self.nc_file))
        regions = gpd.read_file(clisops_test_data["multi_regions_geojson"]).set_index(
            "id"
        )
        regions_utm = regions.to_crs("EPSG:32618")
        assert subset.subset_shape(da, regions).equals(
            subset.subset_shape(da, regions_utm)
        )

    def test_compare_crs(self):
        raster_crs = CRS.from_string("EPSG:4326")
        shape_crs = CRS.from_string("EPSG:32618")
        with pytest.raises(CRSError):
            subset._check_crs_compatibility(raster_crs, shape_crs)


class TestDistance:
    def test_values(self):
        # Check values are OK. Values taken from pyproj test.
        boston_lat = 42.0 + (15.0 / 60.0)
        boston_lon = -71.0 - (7.0 / 60.0)
        portland_lat = 45.0 + (31.0 / 60.0)
        portland_lon = -123.0 - (41.0 / 60.0)

        da = xr.DataArray(
            0, coords={"lon": [boston_lon], "lat": [boston_lat]}, dims=["lon", "lat"]
        )
        d = subset.distance(da, lon=portland_lon, lat=portland_lat)
        np.testing.assert_almost_equal(d, 4164074.239, decimal=3)

    def test_broadcasting(self):
        # Check output dimensions match lons and lats.
        lon = np.linspace(-180, 180, 20)
        lat = np.linspace(-90, 90, 30)
        da = xr.Dataset(data_vars=None, coords={"lon": lon, "lat": lat})
        da["data"] = xr.DataArray(
            np.random.rand(lon.size, lat.size), dims=["lon", "lat"]
        )

        d = subset.distance(da, lon=-34, lat=56).squeeze("site")
        assert d.dims == da.data.dims
        assert d.shape == da.data.shape
        assert d.units == "m"

        # Example of how to get the actual 2D indices.
        k = d.argmin()
        i, j = np.unravel_index(k, da.data.shape)
        assert d[i, j] == d.min()


class TestSubsetLevel:
    nc_plev = "cmip6/o3_Amon_GFDL-ESM4_historical_r1i1p1f1_gr1_185001-194912.nc"
    plevs = [
        100000,
        92500,
        85000,
        70000,
        60000,
        50000,
        40000,
        30000,
        25000,
        20000,
        15000,
        10000,
        7000,
        5000,
        3000,
        2000,
        1000,
        500,
        100,
    ]

    def test_simple(self, nimbus):
        da = xr.open_dataset(nimbus.fetch(self.nc_plev)).o3
        lev_st = 100000.0
        lev_ed = 100.0

        out = subset.subset_level(da, first_level=lev_st, last_level=lev_ed)
        out1 = subset.subset_level(da, first_level=f"{lev_st}", last_level=f"{lev_ed}")
        np.testing.assert_array_equal(out, out1)
        np.testing.assert_array_equal(out.plev.min(), lev_ed)
        np.testing.assert_array_equal(out.plev.max(), lev_st)

    def test_level_outofbounds(self, caplog, nimbus):
        caplog.set_level("WARNING", logger="clisops")
        da = xr.open_dataset(nimbus.fetch(self.nc_plev)).o3
        lev_st = 10000000
        lev_ed = 10

        with ContextLogger(caplog):
            out = subset.subset_level(da, first_level=lev_st, last_level=lev_ed)

        np.testing.assert_array_equal(out.plev.min(), da.plev.min())
        np.testing.assert_array_equal(out.plev.max(), da.plev.max())

        assert (
            '"first_level" has been nudged to nearest valid level in xarray object.'
            in caplog.text
        )
        assert (
            '"last_level" has been nudged to nearest valid level in xarray object.'
            in caplog.text
        )

    def test_warnings(self, caplog, nimbus):
        caplog.set_level("WARNING", logger="clisops")

        da = xr.open_dataset(nimbus.fetch(self.nc_plev)).o3

        with pytest.raises(TypeError):
            subset.subset_level(da, first_level=da, last_level=da)

        with ContextLogger(caplog):
            subset.subset_level(da, first_level="1000", last_level="1000")
            assert (
                '"first_level" should be a number, it has been converted to a float.'
                in caplog.text
            )

            subset.subset_level(da, first_level=81200, last_level=54100.6)

            msgs = {
                '"first_level" has been nudged to nearest valid level in xarray object.',
                '"last_level" has been nudged to nearest valid level in xarray object.',
            }
            assert [m in caplog.text for m in msgs]

    def test_level_first_only(self, nimbus):
        da = xr.open_dataset(nimbus.fetch(self.nc_plev)).o3
        lev_st = 100000

        # first level only
        with warnings.catch_warnings(record=True) as record:
            out = subset.subset_level(da, first_level=lev_st, last_level=lev_st)
        assert not record

        np.testing.assert_array_equal(out.plev.values[0], da.plev.values[0])
        np.testing.assert_array_equal(out.plev.values[-1], da.plev.values[0])

    def test_nudge_levels(self, nimbus):
        da = xr.open_dataset(nimbus.fetch(self.nc_plev)).o3
        # good_levels = [40000, 30000]
        # good_indices = [6, 7]

        with warnings.catch_warnings(record=True) as record:
            out = subset.subset_level(da, first_level=37000.2342, last_level=27777)
        assert not record

        np.testing.assert_array_equal(out.plev.values[0], da.plev.values[7])
        np.testing.assert_array_equal(out.plev.values[-1], da.plev.values[7])

        with warnings.catch_warnings(record=True) as record:
            out = subset.subset_level(da, first_level=41562, last_level=29999)
        assert not record

        np.testing.assert_array_equal(out.plev.values[:], da.plev.values[6:8])


class TestGridPolygon:
    def test_rectilinear(self):
        pytest.importorskip("xesmf", "0.6.2")
        # CF-Compliant with bounds
        ds = xesmf.util.cf_grid_2d(-200, -100, 20, -60, 60, 10)
        poly = subset._rectilinear_grid_exterior_polygon(ds)
        assert poly == Polygon([(-200, -60), (-200, 60), (-100, 60), (-100, -60)])

        # Without bounds
        ds.lon.attrs.pop("bounds")
        ds.lat.attrs.pop("bounds")
        ds = xr.Dataset(dict(lon=ds.lon, lat=ds.lat))
        poly = subset._rectilinear_grid_exterior_polygon(ds)
        assert poly == Polygon([(-200, -60), (-200, 60), (-100, 60), (-100, -60)])

    @pytest.mark.parametrize("mode", ["bbox"])
    def test_curvilinear(self, mode, nimbus):
        """
        Note that there is at least one error in the lat and lon vertices.
        >>> print(ds.vertices_latitude[0, 283].data)
        [-78.29248047  81.48125458 - 78.49452209 - 78.49452209]
        >>> print(ds.vertices_longitude[0, 283].data)
        [357.  73. 356. 356.]
        """
        from shapely.geometry import MultiPoint, Point

        ds = xr.open_dataset(
            nimbus.fetch("cmip6/sic_SImon_CCCma-CanESM5_ssp245_r13i1p2f1_2020.nc")
        )

        poly = subset._curvilinear_grid_exterior_polygon(ds, mode=mode)

        # Check that all grid centroids are within the polygon
        pts = list(map(Point, zip(ds.longitude.data.flat, ds.latitude.data.flat)))
        assert MultiPoint(pts).within(poly)


class TestShapeBboxIndexer:
    def test_rectilinear(self):
        # Create small polygon fitting in one cell.
        pytest.importorskip("xesmf", "0.6.2")
        x, y = -150, 35
        p = Point(x, y)
        ds = xesmf.util.cf_grid_2d(-200, 0, 20, -60, 60, 10)

        # Confirm that after subsetting, the polygon is still entirely within the grid.
        for b in [1, 10, 20]:
            pb = p.buffer(b)
            inds = subset.shape_bbox_indexer(ds, gpd.GeoDataFrame(geometry=[pb]))
            assert pb.within(subset.grid_exterior_polygon(ds.isel(inds)))

    def test_complex_geometries(self):
        # Test with geometries that cannot be simplified to a single polygon using `unary_union`.
        pytest.importorskip("xesmf", "0.6.2")
        import shapely.wkt

        p1 = shapely.wkt.loads(
            "POLYGON((-65.5563 49.257, -64.2166 48.5017, -70.8387 45.2339, -74.6375 44.9993, "
            "-65.5563 49.257))"
        )
        p2 = shapely.wkt.loads(
            "POLYGON ((-58.64 51.2, -78.7115 46.326, -78.1958 62.2551, -64.5341 60.309, -58.64 51.2), "
            "(-78.5687 58.6447, -78.5675 58.646, -78.5762 58.6482, -78.5698 58.6445, -78.5687 58.6447), "
            "(-78.5539 58.6486, -78.5515 58.646, -78.5468 58.6507, -78.5515 58.6511, -78.5539 58.6486), "
            "(-78.5375 58.6513, -78.5353 58.6496, -78.5331 58.6503, -78.5357 58.6512, -78.5375 58.6513), "
            "(-78.517 58.6497, -78.5082 58.6452, -78.5068 58.6456, -78.5112 58.6484, -78.517 58.6497), "
            "(-78.5387 58.6484, -78.5411 58.6509, -78.5466 58.6485, -78.5355 58.6459, -78.5387 58.6484), "
            "(-78.5542 58.6516, -78.5543 58.6539, -78.5614 58.6544, -78.5571 58.6516, -78.5542 58.6516), "
            "(-78.5508 58.6622, -78.561 58.6648, -78.5642 58.664, -78.5559 58.6609, -78.5508 58.6622), "
            "(-78.5814 58.6764, -78.5831 58.675, -78.5802 58.6739, -78.5807 58.6761, -78.5814 58.6764))"
        )

        ds = xesmf.util.cf_grid_2d(-200, 0, 20, 0, 71, 10)
        inds = subset.shape_bbox_indexer(ds, gpd.GeoDataFrame(geometry=[p1, p2]))
        assert "lon" in inds and "lat" in inds, "Expected lon and lat in indexer."
        env = subset.grid_exterior_polygon(ds.isel(inds))
        assert p1.within(env)
        assert p2.within(env)

        # inds should be empty is region is not contained in grid exterior geometry
        ds = xesmf.util.cf_grid_2d(
            -200, 0, 20, 0, 61, 10
        )  # polygon goes up to 62, grid stops at 60
        inds = subset.shape_bbox_indexer(ds, gpd.GeoDataFrame(geometry=[p1, p2]))
        assert inds == {}

    def test_curvilinear(self):
        # Check that grid along lon/lat and a rotated grid are indexed identically for geometry and rotated geometry.
        pytest.importorskip("xesmf", "0.6.2")
        from shapely.affinity import rotate

        ds = xesmf.util.grid_2d(0, 100, 10, 0, 60, 6)
        rds = rotated_grid_2d(0, 100, 10, 0, 60, 6, angle=45)

        geom = Polygon(([0, 0], [50, 0], [50, 30], [0, 30]))
        rgeom = rotate(geom, 45, origin=Point(0, 0))

        # The subsetted grid should have the same dimensions as the subsetted rotated grid.
        i = subset.shape_bbox_indexer(ds, gpd.GeoDataFrame(geometry=[geom]))
        ri = subset.shape_bbox_indexer(rds, gpd.GeoSeries([rgeom]))
        assert ri == i

    def test_multipoints(self):
        # Test with a MultiPoint geometry.
        pytest.importorskip("xesmf", "0.6.2")
        from shapely.geometry import MultiPoint, Point

        ds = xesmf.util.cf_grid_2d(-200, 0, 20, -60, 60, 10)

        coords = (-150, 35), (-100, 40), (-125, 55)

        for n in range(1, 4):
            geom = [Point(coords[i]) for i in range(n)]
            inds = subset.shape_bbox_indexer(ds, gpd.GeoDataFrame(geometry=geom))
            inds2 = subset.shape_bbox_indexer(ds, geom)
            assert inds2 == inds
            inds3 = subset.shape_bbox_indexer(ds, gpd.points_from_xy(*zip(*coords[:n])))
            assert inds3 == inds3
            assert MultiPoint(geom).within(subset.grid_exterior_polygon(ds.isel(inds)))

        for n in range(1, 4):
            geom = MultiPoint([coords[i] for i in range(n)])
            inds = subset.shape_bbox_indexer(ds, gpd.GeoDataFrame(geometry=[geom]))
            assert geom.within(subset.grid_exterior_polygon(ds.isel(inds)))


def rotated_grid_2d(lon0_b, lon1_b, d_lon, lat0_b, lat1_b, d_lat, angle):
    # Rotate lat lon by degree.
    ds = xesmf.util.grid_2d(lon0_b, lon1_b, d_lon, lat0_b, lat1_b, d_lat)

    # Rotation matrix
    theta = np.radians(angle)
    c, s = np.cos(theta), np.sin(theta)
    r = np.array(((c, -s), (s, c)))

    fds = ds.stack(z=("x", "y"), z_b=("x_b", "y_b"))
    fds.lon[:], fds.lat[:] = np.matmul(r, xr.concat([fds.lon, fds.lat], dim="c").data)
    fds.lon_b[:], fds.lat_b[:] = np.matmul(
        r, xr.concat([fds.lon_b, fds.lat_b], dim="c").data
    )

    return fds.unstack(("z", "z_b"))



import functools
import os
import re
import sys
import warnings
from pathlib import Path
from types import FunctionType, ModuleType
from typing import Optional, Union

from dask.utils import byte_sizes
from loguru import logger
from packaging.version import Version


def parse_size(size):
    """

    Parse size string into number of bytes.

    :param size: (str) size to parse in any unit
    :return: (int) number of bytes
    """
    n, suffix = re.match(r"^(\d+\.?\d*)([a-zA-Z]+)$", size).groups()

    try:
        multiplier = byte_sizes[suffix.lower()]

        size_in_bytes = multiplier * float(n)
    except KeyError as err:
        raise ValueError(f"Could not interpret '{suffix}' as a byte unit") from err

    return size_in_bytes


def expand_wildcards(paths: Union[str, Path]) -> list:
    """Expand the wildcards that may be present in Paths."""
    path = Path(paths).expanduser()
    parts = path.parts[1:] if path.is_absolute() else path.parts
    return [f for f in Path(path.root).glob(str(Path("").joinpath(*parts)))]


def require_module(
    func: FunctionType,
    module: ModuleType,
    module_name: str,
    min_version: Optional[str] = "0.0.0",
    max_supported_version: Optional[str] = None,
    max_supported_warning: Optional[str] = None,
):
    """Ensure that module is installed before function/method is called, decorator."""

    @functools.wraps(func)
    def wrapper_func(*args, **kwargs):
        exception_msg = (
            f"Package {module_name} >= {min_version} is required to use {func}."
        )
        if module is None:
            raise ModuleNotFoundError(exception_msg)
        if Version(module.__version__) < Version(min_version):
            raise ImportError(exception_msg)

        if max_supported_version is not None:
            if Version(module.__version__) > Version(max_supported_version):
                if max_supported_warning is not None:
                    warnings.warn(max_supported_warning)
                else:
                    warnings.warn(
                        f"Package {module_name} version {module.__version__} "
                        f"is greater than the suggested version {max_supported_version}."
                    )
        return func(*args, **kwargs)

    return wrapper_func


def check_dir(func: FunctionType, dr: Union[str, Path]):
    """Ensure that directory dr exists before function/method is called, decorator."""
    if not os.path.isdir(dr):
        os.makedirs(dr)

    @functools.wraps(func)
    def wrapper_func(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper_func


def _logging_examples() -> None:
    """Testing module."""
    logger.trace("0")
    logger.debug("1")
    logger.info("2")
    logger.success("2.5")
    logger.warning("3")
    logger.error("4")
    logger.critical("5")


def enable_logging() -> list[int]:
    logger.enable("clisops")

    config = dict(
        handlers=[
            dict(
                sink=sys.stdout,
                format="<green>{time:YYYY-MM-DD HH:mm:ss.SSS Z UTC}</>"
                " <red>|</> <lvl>{level}</> <red>|</> <cyan>{name}:{function}:{line}</>"
                " <red>|</> <lvl>{message}</>",
                level="INFO",
            ),
            dict(
                sink=sys.stderr,
                format="<red>{time:YYYY-MM-DD HH:mm:ss.SSS Z UTC} | {level} | {name}:{function}:{line} | {message}</>",
                level="WARNING",
            ),
        ]
    )
    return logger.configure(**config)


def _list_ten(list1d):
    """Convert list to string of 10 list elements equally distributed to beginning and end of the list.

    Parameters
    ----------
    list1d : list
        1D list.

    Returns
    -------
    str
        String containing the comma separated 5 first and last elements of the list, with "..." in between.
        For example "1, 2, 3, 4, 5 ... , 20, 21, 22, 23, 24, 25".
    """
    if len(list1d) < 11:
        return ", ".join(str(i) for i in list1d)
    else:
        return (
            ", ".join(str(i) for i in list1d[0:5])
            + " ... "
            + ", ".join(str(i) for i in list1d[-5:])
        )



import os

import pytest
import xarray as xr
from packaging.version import Version

from clisops.core.average import XESMF_MINIMUM_VERSION  # noqa
from clisops.exceptions import InvalidParameterValue
from clisops.ops.average import average_over_dims, average_shape, average_time

try:
    import xesmf

    if Version(xesmf.__version__) < Version(XESMF_MINIMUM_VERSION):
        raise ImportError("xesmf version is too old")
except ImportError:
    xesmf = None

XESMF_IMPORT_MESSAGE = f"xesmf >= {XESMF_MINIMUM_VERSION} is needed for average_shape"


def _check_output_nc(result, fname="output_001.nc"):
    assert fname in [os.path.basename(_) for _ in result]


def _load_ds(fpath):
    return xr.open_mfdataset(
        fpath, decode_times=xr.coders.CFDatetimeCoder(use_cftime=True)
    )


def test_average_basic_data_array(nimbus):
    ds = xr.open_dataset(
        nimbus.fetch("cmip5/tas_Amon_HadGEM2-ES_rcp85_r1i1p1_200512-203011.nc")
    )
    result = average_over_dims(
        ds.tas, dims=["time"], ignore_undetected_dims=False, output_type="xarray"
    )
    assert "time" not in result[0]


def test_average_time_xarray(mini_esgf_data):
    result = average_over_dims(
        mini_esgf_data["CMIP5_TAS"],
        dims=["time"],
        ignore_undetected_dims=False,
        output_type="xarray",
    )

    assert "time" not in result[0]


def test_average_lat_xarray(mini_esgf_data):
    result = average_over_dims(
        mini_esgf_data["CMIP5_TAS"],
        dims=["latitude"],
        ignore_undetected_dims=False,
        output_type="xarray",
    )

    assert "lat" not in result[0]


def test_average_lon_xarray(mini_esgf_data):
    result = average_over_dims(
        mini_esgf_data["CMIP5_TAS"],
        dims=["longitude"],
        ignore_undetected_dims=False,
        output_type="xarray",
    )

    assert "lon" not in result[0]


def test_average_level_xarray(nimbus):
    result = average_over_dims(
        nimbus.fetch(
            "cmip6/o3_Amon_GFDL-ESM4_historical_r1i1p1f1_gr1_185001-194912.nc",
        ),
        dims=["level"],
        ignore_undetected_dims=False,
        output_type="xarray",
    )

    assert "plev" not in result[0]


def test_average_time_nc(tmpdir, mini_esgf_data):
    result = average_over_dims(
        mini_esgf_data["CMIP5_TAS"],
        dims=["time"],
        ignore_undetected_dims=False,
        output_dir=tmpdir,
        output_type="netcdf",
        file_namer="standard",
    )
    _check_output_nc(result, fname="tas_mon_HadGEM2-ES_rcp85_r1i1p1_avg-t.nc")


def test_average_lat_nc(tmpdir, mini_esgf_data):
    result = average_over_dims(
        mini_esgf_data["CMIP5_TAS"],
        dims=["latitude"],
        ignore_undetected_dims=False,
        output_dir=tmpdir,
        output_type="netcdf",
        file_namer="standard",
    )

    _check_output_nc(
        result, fname="tas_mon_HadGEM2-ES_rcp85_r1i1p1_20051216-22991216_avg-y.nc"
    )


def test_average_lon_nc(tmpdir, mini_esgf_data):
    result = average_over_dims(
        mini_esgf_data["CMIP5_TAS"],
        dims=["longitude"],
        ignore_undetected_dims=False,
        output_dir=tmpdir,
        output_type="netcdf",
        file_namer="standard",
    )

    _check_output_nc(
        result, fname="tas_mon_HadGEM2-ES_rcp85_r1i1p1_20051216-22991216_avg-x.nc"
    )


def test_average_level_nc(nimbus, tmpdir):
    result = average_over_dims(
        nimbus.fetch(
            "cmip6/o3_Amon_GFDL-ESM4_historical_r1i1p1f1_gr1_185001-194912.nc",
        ),
        dims=["level"],
        ignore_undetected_dims=False,
        output_dir=tmpdir,
        output_type="netcdf",
        file_namer="standard",
    )

    _check_output_nc(
        result,
        fname="o3_Amon_GFDL-ESM4_historical_r1i1p1f1_gr1_18500116-19491216_avg-z.nc",
    )


def test_average_multiple_dims_filename(tmpdir, mini_esgf_data):
    result = average_over_dims(
        mini_esgf_data["CMIP5_TAS"],
        dims=["time", "longitude"],
        ignore_undetected_dims=False,
        output_dir=tmpdir,
        output_type="netcdf",
        file_namer="standard",
    )

    _check_output_nc(result, fname="tas_mon_HadGEM2-ES_rcp85_r1i1p1_avg-tx.nc")


def test_average_multiple_dims_xarray(mini_esgf_data):
    result = average_over_dims(
        mini_esgf_data["CMIP5_TAS"],
        dims=["time", "longitude"],
        ignore_undetected_dims=False,
        output_type="xarray",
    )

    assert "time" not in result[0]
    assert "lon" not in result[0]


def test_average_no_dims(mini_esgf_data):
    with pytest.raises(InvalidParameterValue) as exc:
        average_over_dims(
            mini_esgf_data["CMIP5_TAS"],
            dims=None,
            ignore_undetected_dims=False,
            output_type="xarray",
        )
    assert str(exc.value) == "At least one dimension for averaging must be provided"


def test_unknown_dim(mini_esgf_data):
    with pytest.raises(InvalidParameterValue) as exc:
        average_over_dims(
            mini_esgf_data["CMIP5_TAS"],
            dims=["wrong"],
            ignore_undetected_dims=False,
            output_type="xarray",
        )
    assert (
        str(exc.value)
        == "Dimensions for averaging must be one of ['time', 'level', 'latitude', 'longitude', 'realization']"
    )


def test_dim_not_found(mini_esgf_data):
    with pytest.raises(InvalidParameterValue) as exc:
        average_over_dims(
            mini_esgf_data["CMIP5_TAS"],
            dims=["level", "time"],
            ignore_undetected_dims=False,
            output_type="xarray",
        )
    assert (
        str(exc.value)
        == "Requested dimensions were not found in input dataset: {'level'}."
    )


def test_dim_not_found_ignore(mini_esgf_data):
    # cannot average over level, but have ignored it and done average over time anyway
    result = average_over_dims(
        mini_esgf_data["CMIP5_TAS"],
        dims=["level", "time"],
        ignore_undetected_dims=True,
        output_type="xarray",
    )

    assert "time" not in result[0]
    assert "height" in result[0]


# FIXME: This kind of test is not desirable as it is testing the internal testing implementation
# def test_aux_variables():
#     """
#     test auxiliary variables are remembered in output dataset
#     Have to create a netcdf file with auxiliary variable
#     """
#
#     ds = _load_ds("tests/data/test_file.nc")
#
#     assert "do_i_get_written" in ds.variables
#
#     result = average_over_dims(
#         ds=ds,
#         dims=["level", "time"],
#         ignore_undetected_dims=True,
#         output_type="xarray",
#     )
#
#     assert "do_i_get_written" in result[0].variables


@pytest.mark.skipif(xesmf is None, reason=XESMF_IMPORT_MESSAGE)
def test_average_shape_xarray(mini_esgf_data, clisops_test_data):
    result = average_shape(
        ds=mini_esgf_data["CMIP6_TAS_ONE_TIME_STEP"],
        shape=clisops_test_data["meridian_geojson"],
        variable=None,
        output_type="xarray",
    )
    assert "geom" in result[0]


@pytest.mark.skipif(xesmf is None, reason=XESMF_IMPORT_MESSAGE)
def test_average_multiple_shapes_xarray(mini_esgf_data, clisops_test_data):
    result = average_shape(
        mini_esgf_data["CMIP6_TAS_ONE_TIME_STEP"],
        clisops_test_data["multi_regions_geojson"],
        output_type="xarray",
    )

    assert result[0].geom.size > int(1)


@pytest.mark.skipif(xesmf is None, reason=XESMF_IMPORT_MESSAGE)
def test_average_shape_no_shape(mini_esgf_data):
    # Run without JSON file
    with pytest.raises(InvalidParameterValue) as exc:
        average_shape(
            ds=mini_esgf_data["CMIP6_TAS_ONE_TIME_STEP"],
            shape=None,  # noqa
            variable=None,
            output_type="xarray",
        )
    assert str(exc.value) == "At least one area for averaging must be provided"


@pytest.mark.skipif(xesmf is None, reason=XESMF_IMPORT_MESSAGE)
def test_average_shape_nc(tmpdir, mini_esgf_data, clisops_test_data):
    result = average_shape(
        ds=mini_esgf_data["CMIP6_TAS_ONE_TIME_STEP"],
        shape=clisops_test_data["meridian_geojson"],
        variable=None,
        output_dir=tmpdir,
        output_type="netcdf",
        file_namer="standard",
    )
    _check_output_nc(
        result,
        fname="tas_Amon_FGOALS-g3_historical_r1i1p1f1_gn_18500116-18500116_avg-shape.nc",
    )


def test_average_over_years(mini_esgf_data):
    ds = _load_ds(mini_esgf_data["CMIP5_TAS"])  # monthly dataset

    # check initial dataset
    assert ds.time.shape == (3530,)
    assert ds.time.values[0].isoformat() == "2005-12-16T00:00:00"
    assert ds.time.values[-1].isoformat() == "2299-12-16T00:00:00"

    result = average_time(
        mini_esgf_data["CMIP5_TAS"],
        freq="year",
        output_type="xarray",
    )

    time_length = ds.time.values[-1].year - ds.time.values[0].year + 1
    assert result[0].time.shape == (time_length,)  # get number of years
    assert result[0].time.values[0].isoformat() == "2005-01-01T00:00:00"
    assert result[0].time.values[-1].isoformat() == "2299-01-01T00:00:00"

    # test time bounds
    assert [t.isoformat() for t in result[0].time_bnds.values[0]] == [
        "2005-01-01T00:00:00",
        "2005-12-30T00:00:00",
    ]
    assert [t.isoformat() for t in result[0].time_bnds.values[-1]] == [
        "2299-01-01T00:00:00",
        "2299-12-30T00:00:00",
    ]


def test_average_over_months(mini_esgf_data):
    ds = _load_ds(mini_esgf_data["CMIP6_SICONC_DAY"])  # monthly dataset

    # check initial dataset
    assert ds.time.shape == (60225,)
    assert ds.time.values[0].isoformat() == "1850-01-01T12:00:00"
    assert ds.time.values[-1].isoformat() == "2014-12-31T12:00:00"

    # average over time
    result = average_time(
        mini_esgf_data["CMIP6_SICONC_DAY"],
        freq="month",
        output_type="xarray",
    )

    time_length = (
        ds.time.values[-1].year - ds.time.values[0].year + 1
    ) * 12  # get number of months

    assert result[0].time.shape == (time_length,)
    assert result[0].time.values[0].isoformat() == "1850-01-01T00:00:00"
    assert result[0].time.values[-1].isoformat() == "2014-12-01T00:00:00"

    # test time bounds
    assert [t.isoformat() for t in result[0].time_bnds.values[0]] == [
        "1850-01-01T00:00:00",
        "1850-01-31T00:00:00",
    ]
    assert [t.isoformat() for t in result[0].time_bnds.values[-1]] == [
        "2014-12-01T00:00:00",
        "2014-12-31T00:00:00",
    ]


def test_average_time_no_freq(mini_esgf_data):
    with pytest.raises(InvalidParameterValue) as exc:
        # average over time
        average_time(
            mini_esgf_data["CMIP6_SICONC_DAY"],
            freq=None,  # noqa
            output_type="xarray",
        )
    assert str(exc.value) == "At least one frequency for averaging must be provided"


def test_average_time_incorrect_freq(mini_esgf_data):
    with pytest.raises(InvalidParameterValue) as exc:
        # average over time
        average_time(
            mini_esgf_data["CMIP6_SICONC_DAY"],
            freq="week",
            output_type="xarray",
        )
    assert (
        str(exc.value)
        == "Time frequency for averaging must be one of ['day', 'month', 'year']."
    )


def test_average_time_file_name(tmpdir, mini_esgf_data):
    result = average_time(
        mini_esgf_data["CMIP5_TAS"],
        freq="year",
        output_type="nc",
        output_dir=tmpdir,
    )

    _check_output_nc(
        result, fname="tas_mon_HadGEM2-ES_rcp85_r1i1p1_20050101-22990101_avg-year.nc"
    )


def test_average_time_cordex(mini_esgf_data):
    ds = _load_ds(mini_esgf_data["C3S_CORDEX_EUR_ZG500"])

    # check initial dataset
    assert ds.time.shape == (3653,)
    assert ds.time.values[0].isoformat() == "2071-01-01T12:00:00"
    assert ds.time.values[-1].isoformat() == "2080-12-31T12:00:00"

    # average over time
    result = average_time(
        mini_esgf_data["C3S_CORDEX_EUR_ZG500"],
        freq="month",
        output_type="xarray",
    )

    time_length = (
        ds.time.values[-1].year - ds.time.values[0].year + 1
    ) * 12  # get number of months

    assert result[0].time.shape == (time_length,)
    assert result[0].time.values[0].isoformat() == "2071-01-01T00:00:00"
    assert result[0].time.values[-1].isoformat() == "2080-12-01T00:00:00"

    # test time bounds
    assert [t.isoformat() for t in result[0].time_bnds.values[0]] == [
        "2071-01-01T00:00:00",
        "2071-01-31T00:00:00",
    ]
    assert [t.isoformat() for t in result[0].time_bnds.values[-1]] == [
        "2080-12-01T00:00:00",
        "2080-12-31T00:00:00",
    ]



# This information is located in its own file so that it can be loaded
# without importing the main package when its dependencies are not installed.
# See: https://packaging.python.org/guides/single-sourcing-package-version

__author__ = """Eleanor Smith"""
__contact__ = "eleanor.smith@stfc.ac.uk"
__copyright__ = "Copyright 2018-2024. United Kingdom Research and Innovation"
__license__ = "BSD-3-Clause"
__version__ = "0.16.2"



import pytest
import xarray as xr

from clisops import CONFIG
from clisops.exceptions import InvalidProject
from clisops.ops.subset import subset
from clisops.parameter import time_interval
from clisops.utils.file_namers import get_file_namer


def test_SimpleFileNamer():
    s = get_file_namer("simple")()

    checks = [
        (("my.stuff", "netcdf"), "output_001.nc"),
        (("other", "netcdf"), "output_002.nc"),
    ]

    for args, expected in checks:
        resp = s.get_file_name(*args)
        assert resp == expected


def test_SimpleFileNamer_no_fmt():
    s = get_file_namer("simple")()

    checks = (("my.stuff", None),)

    for args in checks:
        with pytest.raises(KeyError):
            s.get_file_name(*args)


def test_SimpleFileNamer_with_chunking(tmpdir, mini_esgf_data):
    start_time, end_time = "2001-01-01T00:00:00", "2200-12-30T00:00:00"
    area = (0.0, 10.0, 175.0, 90.0)

    config_max_file_size = CONFIG["clisops:write"]["file_size_limit"]
    temp_max_file_size = "10KB"
    CONFIG["clisops:write"]["file_size_limit"] = temp_max_file_size
    outputs = subset(
        ds=mini_esgf_data["CMIP5_TAS"],
        time=time_interval(start_time, end_time),
        area=area,
        output_dir=tmpdir,
        output_type="nc",
        file_namer="simple",
    )

    CONFIG["clisops:write"]["file_size_limit"] = config_max_file_size

    count = 0
    for output in outputs:
        count += 1
        assert f"output_00{count}.nc" in output


def test_StandardFileNamer_no_project_match():
    s = get_file_namer("standard")()

    class Thing:
        pass

    mock_ds = Thing()
    mock_ds.attrs = {}

    with pytest.raises(InvalidProject):
        s.get_file_name(mock_ds)


def test_StandardFileNamer_cmip5(mini_esgf_data):
    s = get_file_namer("standard")()

    _ds = xr.open_mfdataset(
        mini_esgf_data["CMIP5_TAS"],
        decode_times=xr.coders.CFDatetimeCoder(use_cftime=True),
        combine="by_coords",
    )

    checks = [(_ds, "tas_mon_HadGEM2-ES_rcp85_r1i1p1_20051216-22991216.nc")]

    for ds, expected in checks:
        resp = s.get_file_name(ds)
        assert resp == expected


def test_StandardFileNamer_cmip5_use_default_attr_names(mini_esgf_data):
    s = get_file_namer("standard")()

    _ds = xr.open_mfdataset(
        mini_esgf_data["CMIP5_TAS"],
        decode_times=xr.coders.CFDatetimeCoder(use_cftime=True),
        combine="by_coords",
    )

    checks = [(_ds, "tas_mon_no-model_rcp85_r1i1p1_20051216-22991216.nc")]
    del _ds.attrs["model_id"]

    for ds, expected in checks:
        resp = s.get_file_name(ds)
        assert resp == expected


def test_StandardFileNamer_cmip6(mini_esgf_data):
    s = get_file_namer("standard")()

    _ds = xr.open_mfdataset(
        mini_esgf_data["CMIP6_SICONC"],
        decode_times=xr.coders.CFDatetimeCoder(use_cftime=True),
        combine="by_coords",
    )

    checks = [(_ds, "siconc_SImon_CanESM5_historical_r1i1p1f1_gn_18500116-20141216.nc")]

    for ds, expected in checks:
        resp = s.get_file_name(ds)
        assert resp == expected


def test_StandardFileNamer_cmip6_use_default_attr_names(mini_esgf_data):
    s = get_file_namer("standard")()

    _ds = xr.open_mfdataset(
        mini_esgf_data["CMIP6_SICONC"],
        decode_times=xr.coders.CFDatetimeCoder(use_cftime=True),
        combine="by_coords",
    )

    checks = [
        (_ds, "siconc_SImon_no-model_historical_r1i1p1f1_no-grid_18500116-20141216.nc")
    ]
    del _ds.attrs["source_id"]
    del _ds.attrs["grid_label"]

    for ds, expected in checks:
        resp = s.get_file_name(ds)
        assert resp == expected


@pytest.mark.skipif(
    condition="platform.system() == 'Windows'",
    reason="Git modules not working on Windows",
)
def test_StandardFileNamer_c3s_cordex(mini_esgf_data):
    s = get_file_namer("standard")()

    _ds = xr.open_mfdataset(
        mini_esgf_data["C3S_CORDEX_NAM_PR"],
        decode_times=xr.coders.CFDatetimeCoder(use_cftime=True),
        combine="by_coords",
    )

    checks = [
        (
            _ds,
            "pr_NAM-22_NOAA-GFDL-GFDL-ESM2M_rcp45_r1i1p1_OURANOS-CRCM5_v1_day_20510101-20601231.nc",
        )
    ]

    for ds, expected in checks:
        resp = s.get_file_name(ds)
        assert resp == expected


@pytest.mark.skipif(
    condition="platform.system() == 'Windows'",
    reason="Git modules not working on Windows",
)
def test_StandardFileNamer_c3s_cordex_use_default_attr_names(mini_esgf_data):
    s = get_file_namer("standard")()

    _ds = xr.open_mfdataset(
        mini_esgf_data["C3S_CORDEX_NAM_PR"],
        decode_times=xr.coders.CFDatetimeCoder(use_cftime=True),
        combine="by_coords",
    )

    checks = [
        (
            _ds,
            "pr_no-domain_NOAA-GFDL-GFDL-ESM2M_rcp45_rXiXpX_OURANOS-CRCM5_v1_day_20510101-20601231.nc",
        )
    ]
    del _ds.attrs["CORDEX_domain"]
    del _ds.attrs["driving_model_ensemble_member"]

    for ds, expected in checks:
        resp = s.get_file_name(ds)
        assert resp == expected


def test_StandardFileNamer_c3s_atlas_v0(mini_esgf_data):
    """Test C3S ATLAS v0 (c3s-ipcc-ar6-atlas) filenamer"""
    s = get_file_namer("standard")()

    _ds = xr.open_mfdataset(
        mini_esgf_data["ATLAS_v0_CORDEX_NAM"],
        decode_times=xr.coders.CFDatetimeCoder(use_cftime=True),
        combine="by_coords",
    )

    checks = [
        (
            _ds,
            "rx1day_CORDEX-NAM_historical_mon_19700101-20051201.nc",
        )
    ]

    for ds, expected in checks:
        resp = s.get_file_name(ds)
        assert resp == expected


def test_StandardFileNamer_c3s_atlas_v1(mini_esgf_data):
    """Test C3S ATLAS v1 (c3s-cica-atlas) filenamer"""
    s = get_file_namer("standard")()

    _ds = xr.open_mfdataset(
        mini_esgf_data["ATLAS_v1_ERA5"],
        decode_times=xr.coders.CFDatetimeCoder(use_cftime=True),
        combine="by_coords",
    )

    checks = [
        (
            _ds,
            "psl_ERA5_no-expt_mon_19400101-20221201.nc",
        )
    ]

    for ds, expected in checks:
        resp = s.get_file_name(ds)
        assert resp == expected



import geopandas as gpd
import numpy as np
import pytest
import xarray as xr
from packaging.version import Version

from clisops.core import average
from clisops.core.regrid import XESMF_MINIMUM_VERSION
from clisops.exceptions import InvalidParameterValue
from clisops.utils import dataset_utils as xu

try:
    import xesmf

    if Version(xesmf.__version__) < Version(XESMF_MINIMUM_VERSION):
        raise ImportError()
except ImportError:
    xesmf = None


@pytest.mark.skipif(
    xesmf is None,
    reason=f"xesmf >= {XESMF_MINIMUM_VERSION} is needed for average_shape",
)
class TestAverageShape:
    # Fetch remote netcdf files
    nc_file = "cmip5/tas_Amon_CanESM2_rcp85_r1i1p1_200701-200712.nc"
    lons_2d_nc_file = "cmip6/sic_SImon_CCCma-CanESM5_ssp245_r13i1p2f1_2020.nc"
    nc_file_neglons = "NRCANdaily/nrcan_canada_daily_tasmax_1990.nc"

    def test_wraps(self, tmp_netcdf_filename, nimbus, clisops_test_data):
        ds = xr.open_dataset(nimbus.fetch(self.nc_file))

        # xESMF has a problem with averaging over dataset when non-averaged variables are present...
        avg = average.average_shape(ds.tas, clisops_test_data["meridian_geojson"])

        # Check attributes are copied
        assert avg.attrs["units"] == ds.tas.attrs["units"]

        # No time subsetting should occur.
        assert len(avg.time) == 12

        # Average temperature at surface for region in January (time=0)
        np.testing.assert_array_almost_equal(avg.isel(time=0), 285.533, 3)

        # Test with Dataset input
        davg = average.average_shape(ds, clisops_test_data["meridian_geojson"]).tas
        xr.testing.assert_equal(davg, avg)

        # With multiple polygons
        poly = gpd.read_file(clisops_test_data["meridian_multi_geojson"])

        avg = average.average_shape(ds, poly).tas
        np.testing.assert_array_almost_equal(avg.isel(time=0), 280.965, 3)

    def test_no_wraps(self, tmp_netcdf_filename, nimbus, clisops_test_data):
        ds = xr.open_dataset(nimbus.fetch(self.nc_file))

        avg = average.average_shape(ds.tas, clisops_test_data["poslons_geojson"])

        # No time subsetting should occur.
        assert len(avg.time) == 12

        # Average temperature at surface for region in January (time=0)
        np.testing.assert_array_almost_equal(avg.isel(time=0), 276.152, 3)

    def test_all_neglons(self, nimbus, clisops_test_data):
        ds = xr.open_dataset(nimbus.fetch(self.nc_file_neglons))

        avg = average.average_shape(ds.tasmax, clisops_test_data["southern_qc_geojson"])

        # Average temperature at surface for region in January (time=0)
        np.testing.assert_array_almost_equal(avg.isel(time=0), 269.257, 3)

    # 2D lat/lon grids are buggy with current xesmf
    # def test_rotated_pole_with_time(self):
    #     ds = xr.open_dataset(self.lons_2d_nc_file)

    #     avg = average.average_shape(ds.rename(vertices='bounds'), self.eastern_canada_geojson)

    def test_average_multiregions(self, nimbus, clisops_test_data):
        ds = xr.open_dataset(nimbus.fetch(self.nc_file))
        regions = gpd.read_file(clisops_test_data["multi_regions_geojson"]).set_index(
            "id"
        )
        avg = average.average_shape(ds.tas, shape=regions)
        np.testing.assert_array_almost_equal(
            avg.isel(time=0), [268.620, 278.290, 277.863], decimal=3
        )
        np.testing.assert_array_equal(avg.geom, ["Québec", "Europe", "Newfoundland"])

    def test_non_overlapping_regions(self, nimbus, clisops_test_data):
        ds = xr.open_dataset(nimbus.fetch(self.nc_file_neglons))
        regions = gpd.read_file(clisops_test_data["meridian_geojson"])

        with pytest.raises(ValueError):
            average.average_shape(ds.tasmax, shape=regions)


class TestAverageOverDims:
    nc_file = "cmip5/tas_Amon_CanESM2_rcp85_r1i1p1_200701-200712.nc"

    def test_average_no_dims(self, nimbus):
        ds = xr.open_dataset(nimbus.fetch(self.nc_file))

        with pytest.raises(InvalidParameterValue) as exc:
            average.average_over_dims(ds)
        assert str(exc.value) == "At least one dimension for averaging must be provided"

    def test_average_one_dim(self, nimbus):
        ds = xr.open_dataset(nimbus.fetch(self.nc_file))

        avg_ds = average.average_over_dims(ds, ["latitude"])

        # time has not been averaged
        assert len(avg_ds.time) == 12

        # lat has been averaged over
        assert "lat" not in avg_ds.dims

    def test_average_two_dims(self, nimbus):
        ds = xr.open_dataset(nimbus.fetch(self.nc_file))

        avg_ds = average.average_over_dims(ds, ["latitude", "time"])

        # time has been averaged over
        assert "time" not in avg_ds.dims

        # lat has been averaged over
        assert "lat" not in avg_ds.dims

    def test_average_wrong_dim(self, nimbus):
        ds = xr.open_dataset(nimbus.fetch(self.nc_file))

        with pytest.raises(InvalidParameterValue) as exc:
            average.average_over_dims(ds, ["wrong", "latitude"])
        assert (
            str(exc.value)
            == "Dimensions for averaging must be one of ['time', 'level', 'latitude', 'longitude', 'realization']"
        )

    def test_undetected_dim(self, nimbus):
        ds = xr.open_dataset(nimbus.fetch(self.nc_file))

        with pytest.raises(InvalidParameterValue) as exc:
            average.average_over_dims(ds, ["level", "time"])
        assert (
            str(exc.value)
            == "Requested dimensions were not found in input dataset: {'level'}."
        )

    def test_average_undetected_dim_ignore(self, nimbus):
        ds = xr.open_dataset(nimbus.fetch(self.nc_file))

        # exception should not be raised as ignore_undetected_dims set to True
        avg_ds = average.average_over_dims(
            ds, ["level", "time"], ignore_undetected_dims=True
        )

        # time has been averaged over
        assert "time" not in avg_ds.dims

    def test_average_wrong_format(self, nimbus):
        ds = xr.open_dataset(nimbus.fetch(self.nc_file))

        with pytest.raises(InvalidParameterValue) as exc:
            average.average_over_dims(ds, [0, "time"])
        assert (
            str(exc.value)
            == "Dimensions for averaging must be one of ['time', 'level', 'latitude', 'longitude', 'realization']"
        )


class TestAverageTime:
    month_ds = "CMIP5_RH"

    def test_average_month(self, mini_esgf_data):
        ds = xu.open_xr_dataset(mini_esgf_data["CMIP6_SICONC_DAY"])
        assert ds.time.shape == (60225,)

        avg_ds = average.average_time(ds, freq="month")
        assert avg_ds.time.shape == (1980,)

    def test_average_year(self, mini_esgf_data):
        ds = xu.open_xr_dataset(mini_esgf_data[self.month_ds])
        assert ds.time.shape == (1752,)

        avg_ds = average.average_time(ds, freq="year")
        assert avg_ds.time.shape == (147,)

    def test_no_freq(self, mini_esgf_data):
        ds = xu.open_xr_dataset(mini_esgf_data[self.month_ds])

        with pytest.raises(InvalidParameterValue) as exc:
            average.average_time(ds, freq=None)  # noqa

        assert str(exc.value) == "At least one frequency for averaging must be provided"

    def test_incorrect_freq(self, mini_esgf_data):
        ds = xu.open_xr_dataset(mini_esgf_data[self.month_ds])

        with pytest.raises(InvalidParameterValue) as exc:
            average.average_time(ds, freq="wrong")
        assert (
            str(exc.value)
            == "Time frequency for averaging must be one of ['day', 'month', 'year']."
        )

    def test_freq_wrong_format(self, mini_esgf_data):
        ds = xu.open_xr_dataset(mini_esgf_data[self.month_ds])

        with pytest.raises(InvalidParameterValue) as exc:
            average.average_time(ds, freq=0)  # noqa

        assert str(exc.value) == "At least one frequency for averaging must be provided"

    def test_no_time(self, mini_esgf_data):
        ds = xu.open_xr_dataset(mini_esgf_data[self.month_ds])
        ds = ds.drop_dims("time")

        with pytest.raises(Exception) as exc:
            average.average_time(ds, freq="year")
        assert str(exc.value) == "Time dimension could not be found"



import re

import numpy as np


def to_isoformat(tm):
    """
    Returns an ISO 8601 string from a time object (of different types).

    :param tm: Time object
    :return: (str) ISO 8601 time string
    """
    if isinstance(tm, np.datetime64):
        return str(tm).split(".")[0]
    else:
        return tm.isoformat()


class AnyCalendarDateTime:
    """
    A class to represent a datetime that could be of any calendar.

    Has the ability to add and subtract a day from the input based on MAX_DAY, MIN_DAY, MAX_MONTH and MIN_MONTH
    """

    MONTH_RANGE = range(1, 13)
    # 31 is the maximum number of days in any month in any of the calendars supported by cftime
    DAY_RANGE = range(1, 32)
    HOUR_RANGE = range(0, 24)
    MINUTE_RANGE = range(0, 60)
    SECOND_RANGE = range(0, 60)

    def __init__(self, year, month, day, hour, minute, second):
        self.year = year

        self.month = month
        self.validate_input(self.month, "month", self.MONTH_RANGE)

        self.day = day
        self.validate_input(self.day, "day", self.DAY_RANGE)

        self.hour = hour
        self.validate_input(self.hour, "hour", self.HOUR_RANGE)

        self.minute = minute
        self.validate_input(self.minute, "minute", self.MINUTE_RANGE)

        self.second = second
        self.validate_input(self.second, "second", self.SECOND_RANGE)

    def validate_input(self, input, name, range):
        if input not in range:
            raise ValueError(
                f"Invalid input {input} for {name}. Expected value between {range[0]} and {range[-1]}."
            )

    def __repr__(self):
        return self.value

    @property
    def value(self):
        return (
            f"{self.year}-{self.month:02d}-{self.day:02d}"
            f"T{self.hour:02d}:{self.minute:02d}:{self.second:02d}"
        )

    def add_day(self):
        """
        Add a day to the input datetime.
        """
        self.day += 1

        if self.day > self.DAY_RANGE[-1]:
            self.month += 1
            self.day = 1

        if self.month > self.MONTH_RANGE[-1]:
            self.year += 1
            self.month = self.MONTH_RANGE[0]

    def sub_day(self, n=1):
        """
        Subtract a day to the input datetime.
        """
        self.day -= 1

        if self.day < self.DAY_RANGE[0]:
            self.month -= 1
            self.day = self.DAY_RANGE[-1]

        if self.month < self.MONTH_RANGE[0]:
            self.year -= 1
            self.month = self.MONTH_RANGE[-1]


def str_to_AnyCalendarDateTime(dt, defaults=None):
    """
    Takes a string representing date/time and returns a DateTimeAnyTime object.
    String formats should start with Year and go through to Second, but you
    can miss out anything from month onwards.

    :param dt: (str) string representing a date/time.
    :param defaults: (list) The default values to use for year, month, day, hour, minute and second if they cannot be parsed from the string. A default value must be provided for each component. If defaults=None, [-1, 1, 1, 0, 0, 0] is used.
    :return: AnyCalendarDateTime object
    """
    if not dt and not defaults:
        raise Exception(
            "Must provide at least the year as argument, or all defaults, to create date time."
        )

    # Start with most common pattern
    regex = re.compile(r"^(\d+)-(\d+)-(\d+)[T ](\d+):(\d+):(\d+)$")
    match = regex.match(dt)

    if match:
        items = match.groups()
    else:
        # Try a more complex split and build of the time string
        if not defaults:
            defaults = [-1, 1, 1, 0, 0, 0]
        else:
            if len(defaults) < 6:
                raise Exception(
                    "A default value must be provided for year, month, day, hour, minute and second."
                )
        components = re.split("[- T:]", dt.strip("Z"))

        # Build a list of time components
        items = components + defaults[len(components) :]

    return AnyCalendarDateTime(*[int(float(i)) for i in items])


def create_time_bounds(ds, freq):
    """Generate time bounds for datasets that have been temporally averaged.

    Averaging frequencies supported are yearly, monthly and daily.
    """
    # get datetime class
    dt_cls = ds.time.values[0].__class__

    if freq == "month":
        time_bounds = [
            [
                dt_cls(tm.year, tm.month, tm.day),
                dt_cls(tm.year, tm.month, tm.daysinmonth),
            ]
            for tm in ds.time.values
        ]

    elif freq == "year":
        # get number of days in december for calendar
        dec_days = dt_cls(2000, 12, 1).daysinmonth
        # generate time bounds
        time_bounds = [
            [dt_cls(tm.year, 1, 1), dt_cls(tm.year, 12, dec_days)]
            for tm in ds.time.values
        ]

    elif freq == "day":
        time_bounds = [
            [
                dt_cls(tm.year, tm.month, tm.day, 0, 0, 0),
                dt_cls(tm.year, tm.month, tm.day, 23, 59, 59),
            ]
            for tm in ds.time.values
        ]

    else:
        raise Exception("Time frequency not supported for creation of time bounds.")

    return time_bounds



import datetime

from clisops.exceptions import InvalidParameterValue
from clisops.parameter._utils import parse_datetime
from clisops.parameter.base_parameter import _BaseIntervalOrSeriesParameter


class TimeParameter(_BaseIntervalOrSeriesParameter):
    """
    Class for time parameter used in subsetting operation.

    | Time can be input as:
    | A string of slash separated values: "2085-01-01T12:00:00Z/2120-12-30T12:00:00Z"
    | A sequence of strings: e.g. ("2085-01-01T12:00:00Z", "2120-12-30T12:00:00Z")

    A time input must be 2 values.

    If using a string input a trailing slash indicates you want to use the earliest/
    latest time of the dataset. e.g. "2085-01-01T12:00:00Z/" will subset from 01/01/2085 to the final time in
    the dataset.

    Validates the times input and parses the values into isoformat.

    """

    def _parse_as_interval(self):
        start, end = self.input.value

        try:
            if start is not None:
                start = parse_datetime(
                    start, defaults=[datetime.MINYEAR, 1, 1, 0, 0, 0]
                )
            if end is not None:
                end = parse_datetime(
                    end, defaults=[datetime.MAXYEAR, 12, 31, 23, 59, 59]
                )

        except Exception:
            raise InvalidParameterValue("Unable to parse the time values entered")

        # Set as None if no start or end, otherwise set as tuple
        value = (start, end)

        if set(value) == {None}:
            value = None

        return value

    def _parse_as_series(self):
        try:
            value = [parse_datetime(tm) for tm in self.input.value]
        except Exception:
            raise InvalidParameterValue("Unable to parse the time values entered")

        return value

    def asdict(self):
        """Returns a dictionary of the time values"""
        if self.type in ("interval", "none"):
            value = self._value_as_tuple()
            return {"start_time": value[0], "end_time": value[1]}
        elif self.type == "series":
            return {"time_values": self.value}

    def get_bounds(self):
        """Returns a tuple of the (start, end) times, calculated from
        the value of the parameter. Either will default to None."""
        if self.type in ("interval", "none"):
            return self._value_as_tuple()

        elif self.type == "series":
            return self.value[0], self.value[-1]

    def __str__(self):
        if self.type in ("interval", "none"):
            value = self._value_as_tuple()
            return (
                f"Time period to subset over"
                f"\n start time: {value[0]}"
                f"\n end time: {value[1]}"
            )
        else:
            return f"Time values to select: {self.value}"



from typing import Optional, Union

import xarray as xr

from clisops import CONFIG
from clisops.project_utils import get_project_name
from clisops.utils.dataset_utils import get_main_variable
from clisops.utils.output_utils import get_format_extension


def get_file_namer(name):
    """Returns the correct filenamer from the provided name."""
    file_namer = {"standard": StandardFileNamer, "simple": SimpleFileNamer}

    return file_namer.get(name, StandardFileNamer)


class _BaseFileNamer:
    """File namer base class."""

    def __init__(self, replace=None, extra=""):
        self._count = 0
        self._replace = replace
        self._extra = extra

    def get_file_name(self, ds, fmt="nc"):
        """Generate numbered file names"""
        self._count += 1
        extension = get_format_extension(fmt)
        return f"output_{self._count:03d}.{extension}"


class SimpleFileNamer(_BaseFileNamer):
    """Simple file namer class.

    Generates numbered file names.
    """

    pass


class StandardFileNamer(SimpleFileNamer):
    """Standard file namer class.

    Generates file names based on input dataset.
    """

    @staticmethod
    def _get_project(ds):
        """Gets the project name from the input dataset"""

        return get_project_name(ds)

    def get_file_name(self, ds, fmt="nc"):
        """Constructs file name."""
        template = self._get_template(ds)

        if not template:
            # Default to parent class namer if no method found
            return super().get_file_name(ds, fmt)

        self._count += 1

        attr_defaults = CONFIG[f"project:{self._get_project(ds)}"]["attr_defaults"]
        attrs = attr_defaults.copy()
        attrs.update(ds.attrs)

        self._resolve_derived_attrs(ds, attrs, template, fmt=fmt)
        file_name = template.format(**attrs)

        return file_name

    def _get_template(self, ds):
        """Gets template to use for output file name, based on the project of the dataset."""
        try:
            return CONFIG[f"project:{self._get_project(ds)}"]["file_name_template"]
        except KeyError:
            return None

    def _resolve_derived_attrs(
        self,
        ds: Union[xr.DataArray, xr.Dataset],
        attrs: dict,
        template: dict,
        fmt: Optional[str] = None,
    ) -> None:
        """Finds var_id, time_range and format_extension of dataset and output to generate output file name."""
        if "__derive__var_id" in template:
            attrs["__derive__var_id"] = get_main_variable(ds)

        if "__derive__time_range" in template:
            attrs["__derive__time_range"] = self._get_time_range(ds)

        if "__derive__extension" in template:
            attrs["__derive__extension"] = get_format_extension(fmt)

        if "extra" in template:
            attrs["extra"] = self._extra

        if self._replace:
            for key, value in self._replace.items():
                attrs[key] = value

    @staticmethod
    def _get_time_range(da: Union[xr.DataArray, xr.Dataset]) -> str:
        """Finds the time range of the data in the output."""
        try:
            times = da.time.values
            return f"_{times.min().strftime('%Y%m%d')}-{times.max().strftime('%Y%m%d')}"
        # catch where "time" attribute cannot be accessed in ds
        except AttributeError:
            return ""



from pathlib import Path
from typing import List, Optional, Union

import xarray as xr
from loguru import logger

from clisops.core import (
    subset_bbox,
    subset_level,
    subset_level_by_values,
    subset_time,
    subset_time_by_components,
    subset_time_by_values,
)
from clisops.core.subset import assign_bounds, get_lat, get_lon  # noqa
from clisops.ops.base_operation import Operation
from clisops.parameter import (
    Interval,
    LevelParameter,
    Series,
    TimeComponents,
    TimeComponentsParameter,
    TimeParameter,
    parameterise,
)
from clisops.parameter.area_parameter import AreaParameter
from clisops.utils.dataset_utils import cf_convert_between_lon_frames

__all__ = ["Subset", "subset"]


class Subset(Operation):
    def _resolve_params(self, **params):
        """Generates a dictionary of subset parameters."""
        time = params.get("time", None)
        area = params.get("area", None)
        level = params.get("level", None)
        time_comps = params.get("time_components", None)

        logger.debug(
            f"Mapping parameters: time: {time}, area: {area}, "
            f"level: {level}, time_components: {time_comps}."
        )

        # Set up args dictionary to be used by `self._calculate()`
        args = dict()

        parameters = parameterise(
            collection=self.ds,
            time=time,
            area=area,
            level=level,
            time_components=time_comps,
        )

        # For each required parameter, check if the parameter can be accessed as a tuple
        # If not: then use the dictionary representation for it
        for param_name in ["time", "area", "level", "time_components"]:
            param_value = parameters.get(param_name)
            if param_value.value is not None:
                args.update(param_value.asdict())

        # Rename start_time and end_time to start_date and end_date to
        # match clisops.core.subset function parameters.
        if "start_time" in args:
            args["start_date"] = args.pop("start_time")

        if "end_time" in args:
            args["end_date"] = args.pop("end_time")

        self.params = args

    def _calculate(self):
        if "lon_bnds" and "lat_bnds" in self.params:
            lon = get_lon(self.ds)
            lat = get_lat(self.ds)

            # ensure lat/lon bounds are in the same order as data, before trying to roll
            # if descending in dataset, they will be flipped in subset_bbox
            self.params["lon_bnds"], self.params["lat_bnds"] = (
                assign_bounds(self.params.get("lon_bnds"), self.ds[lon.name]),
                assign_bounds(self.params.get("lat_bnds"), self.ds[lat.name]),
            )

            # subset with space and optionally time and level
            logger.debug(f"subset_bbox with parameters: {self.params}")
            # bounds are always ascending, so if lon is descending rolling will not work.
            ds, lb, ub = cf_convert_between_lon_frames(
                self.ds, self.params.get("lon_bnds")
            )
            self.params["lon_bnds"] = (lb, ub)
            try:
                kwargs = {}
                valid_args = [
                    "lon_bnds",
                    "lat_bnds",
                    "start_date",
                    "end_date",
                    "first_level",
                    "last_level",
                    "time_values",
                    "level_values",
                ]
                for arg in valid_args:
                    kwargs.setdefault(arg, self.params.get(arg, None))
                result = subset_bbox(ds, **kwargs)
            except NotImplementedError:
                lon_min, lon_max = lon.values.min(), lon.values.max()
                raise Exception(
                    f"The requested longitude subset {self.params.get('lon_bnds')} is not within the longitude bounds "
                    "of this dataset and the data could not be converted to this longitude frame successfully. "
                    "Please re-run your request with longitudes within the bounds of the dataset: "
                    f"({lon_min:.2f}, {lon_max:.2f})"
                )
        else:
            kwargs = {}
            valid_args = ["start_date", "end_date"]
            for arg in valid_args:
                kwargs.setdefault(arg, self.params.get(arg, None))

            # Subset over time interval if requested
            if any(kwargs.values()):
                logger.debug(f"subset_time with parameters: {kwargs}")
                result = subset_time(self.ds, **kwargs)
            # Subset a series of time values if requested
            elif self.params.get("time_values"):
                result = subset_time_by_values(
                    self.ds, time_values=self.params["time_values"]
                )
            else:
                result = self.ds

            # Now test for level subsetting
            kwargs = {}
            valid_args = ["first_level", "last_level"]

            for arg in valid_args:
                kwargs.setdefault(arg, self.params.get(arg, None))

            # Subset with level only
            if any(kwargs.values()):
                # ensure bounds are ascending
                if self.params.get("first_level") > self.params.get("last_level"):
                    first, last = self.params.get("first_level"), self.params.get(
                        "last_level"
                    )
                    self.params["first_level"], self.params["last_level"] = last, first

                logger.debug(f"subset_level with parameters: {kwargs}")
                result = subset_level(result, **kwargs)

            elif self.params.get("level_values", None):
                kwargs = {"level_values": self.params["level_values"]}
                logger.debug(f"subset_level_by_values with parameters: {kwargs}")
                result = subset_level_by_values(result, **kwargs)

        # Now apply time components if specified
        time_comps = self.params.get("time_components")
        if time_comps:
            logger.debug(f"subset_by_time_components with parameters: {time_comps}")
            result = subset_time_by_components(result, time_components=time_comps)

        return result


def subset(
    ds: Union[xr.Dataset, str, Path],
    *,
    time: Optional[Union[str, tuple[str, str], TimeParameter, Series, Interval]] = None,
    area: Optional[
        Union[
            str,
            tuple[
                Union[int, float, str],
                Union[int, float, str],
                Union[int, float, str],
                Union[int, float, str],
            ],
            AreaParameter,
        ]
    ] = None,
    level: Optional[
        Union[
            str,
            tuple[Union[int, float, str], Union[int, float, str]],
            LevelParameter,
            Interval,
        ]
    ] = None,
    time_components: Optional[
        Union[str, dict, TimeComponents, TimeComponentsParameter]
    ] = None,
    output_dir: Optional[Union[str, Path]] = None,
    output_type="netcdf",
    split_method="time:auto",
    file_namer="standard",
) -> list[Union[xr.Dataset, str]]:
    """Subset operation.

    Parameters
    ----------
    ds : Union[xr.Dataset, str]
    time : Optional[Union[str, Tuple[str, str], TimeParameter, Series, Interval]] = None,
    area : str or AreaParameter or Tuple[Union[int, float, str], Union[int, float, str], Union[int, float, str], Union[int, float, str]], optional
    level : Optional[Union[str, Tuple[Union[int, float, str], Union[int, float, str]], LevelParameter, Interval] = None,
    time_components : Optional[Union[str, Dict, TimeComponentsParameter]] = None,
    output_dir : Optional[Union[str, Path]] = None
    output_type : {"netcdf", "nc", "zarr", "xarray"}
    split_method : {"time:auto"}
    file_namer : {"standard", "simple"}

    Returns
    -------
    List[Union[xr.Dataset, str]]
        A list of the subsetted outputs in the format selected; str corresponds to file paths if the
        output format selected is a file.

    Examples
    --------
    | ds: xarray Dataset or "cmip5.output1.MOHC.HadGEM2-ES.rcp85.mon.atmos.Amon.r1i1p1.latest.tas"
    | time: ("1999-01-01T00:00:00", "2100-12-30T00:00:00") or "2085-01-01T12:00:00Z/2120-12-30T12:00:00Z"
    | area: (-5.,49.,10.,65) or "0.,49.,10.,65" or [0, 49.5, 10, 65] with the order being lon_0, lat_0, lon_1, lat_1
    | level: (1000.,) or "1000/2000" or ("1000.50", "2000.60")
    | time_components: "year:2000,2004,2008|month:01,02" or {"year": (2000, 2004, 2008), "months": (1, 2)}
    | output_dir: "/cache/wps/procs/req0111"
    | output_type: "netcdf"
    | split_method: "time:auto"
    | file_namer: "standard"

    Note
    ----
    If you request a selection range (such as level, latitude or longitude) that specifies the lower
    and upper bounds in the opposite direction to the actual coordinate values then clisops.ops.subset
    will detect this issue and reverse your selection before returning the data subset.
    """
    op = Subset(**locals())
    return op.process()



from clisops.exceptions import InvalidParameterValue
from clisops.parameter._utils import to_float
from clisops.parameter.base_parameter import _BaseIntervalOrSeriesParameter


class LevelParameter(_BaseIntervalOrSeriesParameter):
    """
    Class for level parameter used in subsetting operation.

    | Level can be input as:
    | A string of slash separated values: "1000/2000"
    | A sequence of strings: e.g. ("1000.50", "2000.60")
    | A sequence of numbers: e.g. (1000.50, 2000.60)

    A level input must be 2 values.

    If using a string input a trailing slash indicates you want to use the lowest/highest
    level of the dataset. e.g. "/2000" will subset from the lowest level in the dataset
    to 2000.

    Validates the level input and parses the values into numbers.

    """

    def _parse_as_interval(self):
        try:
            value = tuple([to_float(i) for i in self.input.value])
        except InvalidParameterValue:
            raise
        except Exception:
            raise InvalidParameterValue("Unable to parse the level values entered")

        if set(value) == {None}:
            value = None

        return value

    def _parse_as_series(self):
        try:
            value = [to_float(i) for i in self.input.value if i is not None]
        except InvalidParameterValue:
            raise
        except Exception:
            raise InvalidParameterValue("Unable to parse the level values entered")
        return value

    def asdict(self):
        """Returns a dictionary of the level values"""
        if self.type in ("interval", "none"):
            value = self._value_as_tuple()
            return {"first_level": value[0], "last_level": value[1]}
        elif self.type == "series":
            return {"level_values": self.value}

    def __str__(self):
        if self.type in ("interval", "none"):
            value = self._value_as_tuple()
            return (
                f"Level range to subset over"
                f"\n first_level: {value[0]}"
                f"\n last_level: {value[1]}"
            )
        else:
            return f"Level values to select: {self.value}"
